@@ -4,6 +4,7 @@ const Clutter = imports.gi.Clutter;
 const Gdk = imports.gi.Gdk;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
+const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const St = imports.gi.St;
@@ -120,7 +121,7 @@ AltTabPopup.prototype = {
         }
     },
 
-    show : function(backward) {
+    show : function(backward, switch_group) {
         let tracker = Shell.WindowTracker.get_default();
         let apps = tracker.get_running_apps ('');
 
@@ -145,7 +146,16 @@ AltTabPopup.prototype = {
         this._appIcons = this._appSwitcher.icons;
 
         // Make the initial selection
-        if (this._appIcons.length == 1) {
+        if (switch_group) {
+            if (backward) {
+                this._select(0, this._appIcons[0].cachedWindows.length - 1);
+            } else {
+                if (this._appIcons[0].cachedWindows.length > 1)
+                    this._select(0, 1);
+                else
+                    this._select(0, 0);
+            }
+        } else if (this._appIcons.length == 1) {
             if (!backward && this._appIcons[0].cachedWindows.length > 1) {
                 // For compatibility with the multi-app case below
                 this._select(0, 1, true);
@@ -217,41 +227,38 @@ AltTabPopup.prototype = {
 
     _keyPressEvent : function(actor, event) {
         let keysym = event.get_key_symbol();
-        let shift = (Shell.get_event_state(event) & Clutter.ModifierType.SHIFT_MASK);
-        // X allows servers to represent Shift+Tab in two different ways
-        if (shift && keysym == Clutter.Tab)
-            keysym = Clutter.ISO_Left_Tab;
+        let event_state = Shell.get_event_state(event);
+        let backwards = event_state & Clutter.ModifierType.SHIFT_MASK;
+        let action = global.screen.get_display().get_keybinding_action(event.get_key_code(), event_state);
 
         this._disableHover();
 
-        if (keysym == Clutter.grave)
-            this._select(this._currentApp, this._nextWindow());
-        else if (keysym == Clutter.asciitilde)
-            this._select(this._currentApp, this._previousWindow());
+        if (action == Meta.KeyBindingAction.SWITCH_GROUP)
+            this._select(this._currentApp, backwards ? this._previousWindow() : this._nextWindow());
         else if (keysym == Clutter.Escape)
             this.destroy();
         else if (this._thumbnailsFocused) {
-            if (keysym == Clutter.Tab) {
-                if (this._currentWindow == this._appIcons[this._currentApp].cachedWindows.length - 1)
-                    this._select(this._nextApp());
-                else
-                    this._select(this._currentApp, this._nextWindow());
-            } else if (keysym == Clutter.ISO_Left_Tab) {
-                if (this._currentWindow == 0 || this._currentWindow == -1)
-                    this._select(this._previousApp());
-                else
-                    this._select(this._currentApp, this._previousWindow());
-            } else if (keysym == Clutter.Left)
+            if (action == Meta.KeyBindingAction.SWITCH_WINDOWS)
+                if (backwards) {
+                    if (this._currentWindow == 0 || this._currentWindow == -1)
+                        this._select(this._previousApp());
+                    else
+                        this._select(this._currentApp, this._previousWindow());
+                } else {
+                    if (this._currentWindow == this._appIcons[this._currentApp].cachedWindows.length - 1)
+                        this._select(this._nextApp());
+                    else
+                        this._select(this._currentApp, this._nextWindow());
+                }
+            else if (keysym == Clutter.Left)
                 this._select(this._currentApp, this._previousWindow());
             else if (keysym == Clutter.Right)
                 this._select(this._currentApp, this._nextWindow());
             else if (keysym == Clutter.Up)
                 this._select(this._currentApp, null, true);
         } else {
-            if (keysym == Clutter.Tab)
-                this._select(this._nextApp());
-            else if (keysym == Clutter.ISO_Left_Tab)
-                this._select(this._previousApp());
+            if (action == Meta.KeyBindingAction.SWITCH_WINDOWS)
+                this._select(backwards ? this._previousApp() : this._nextApp());
             else if (keysym == Clutter.Left)
                 this._select(this._previousApp());
             else if (keysym == Clutter.Right)
@@ -527,16 +534,11 @@ SwitcherList.prototype = {
         this._leftArrow = new St.DrawingArea({ style_class: 'switcher-arrow',
                                                pseudo_class: 'highlighted' });
         this._leftArrow.connect('repaint', Lang.bind(this,
-                                            function (area) {
-                                                Shell.draw_box_pointer(area, Shell.PointerDirection.LEFT);
-                                            }));
-
+            function() { _drawArrow(this._leftArrow, St.Side.LEFT); }));
         this._rightArrow = new St.DrawingArea({ style_class: 'switcher-arrow',
                                                 pseudo_class: 'highlighted' });
         this._rightArrow.connect('repaint', Lang.bind(this,
-                                            function (area) {
-                                                Shell.draw_box_pointer(area, Shell.PointerDirection.RIGHT);
-                                            }));
+            function() { _drawArrow(this._rightArrow, St.Side.RIGHT); }));
 
         this.actor.add_actor(this._leftArrow);
         this.actor.add_actor(this._rightArrow);
@@ -593,8 +595,8 @@ SwitcherList.prototype = {
     },
 
     addItem : function(item) {
-        let bbox = new St.Clickable({ style_class: 'item-box',
-                                      reactive: true });
+        let bbox = new St.Button({ style_class: 'item-box',
+                                   reactive: true });
 
         bbox.set_child(item);
         this._list.add_actor(bbox);
@@ -992,10 +994,7 @@ AppSwitcher.prototype = {
 
         let n = this._arrows.length;
         let arrow = new St.DrawingArea({ style_class: 'switcher-arrow' });
-        arrow.connect('repaint', Lang.bind(this,
-            function (area) {
-                Shell.draw_box_pointer(area, Shell.PointerDirection.DOWN);
-            }));
+        arrow.connect('repaint', function() { _drawArrow(arrow, St.Side.BOTTOM); });
         this._list.add_actor(arrow);
         this._arrows.push(arrow);
 
@@ -1081,6 +1080,9 @@ ThumbnailList.prototype = {
 
         for (let i = 0; i < this._thumbnailBins.length; i++) {
             let mutterWindow = this._windows[i].get_compositor_private();
+            if (!mutterWindow)
+                continue;
+
             let windowTexture = mutterWindow.get_texture ();
             let [width, height] = windowTexture.get_size();
             let scale = Math.min(1.0, THUMBNAIL_DEFAULT_SIZE / width, availHeight / height);
@@ -1098,3 +1100,46 @@ ThumbnailList.prototype = {
         this._thumbnailBins = new Array();
     }
 };
+
+function _drawArrow(area, side) {
+    let themeNode = area.get_theme_node();
+    let borderColor = themeNode.get_border_color(side);
+    let bodyColor = themeNode.get_foreground_color();
+
+    let [width, height] = area.get_surface_size ();
+    let cr = area.get_context();
+
+    cr.setLineWidth(1.0);
+    Clutter.cairo_set_source_color(cr, borderColor);
+
+    switch (side) {
+    case St.Side.TOP:
+        cr.moveTo(0, height);
+        cr.lineTo(Math.floor(width * 0.5), 0);
+        cr.lineTo(width, height);
+        break;
+
+    case St.Side.BOTTOM:
+        cr.moveTo(width, 0);
+        cr.lineTo(Math.floor(width * 0.5), height);
+        cr.lineTo(0, 0);
+        break;
+
+    case St.Side.LEFT:
+        cr.moveTo(width, height);
+        cr.lineTo(0, Math.floor(height * 0.5));
+        cr.lineTo(width, 0);
+        break;
+
+    case St.Side.RIGHT:
+        cr.moveTo(0, 0);
+        cr.lineTo(width, Math.floor(height * 0.5));
+        cr.lineTo(0, height);
+        break;
+    }
+
+    cr.strokePreserve();
+
+    Clutter.cairo_set_source_color(cr, bodyColor);
+    cr.fill();
+}
