@@ -51,6 +51,9 @@ AlphabeticalView.prototype = {
                 let adjustment = this.actor.vscroll.adjustment;
                 let direction = Overview.SwipeScrollDirection.VERTICAL;
                 Main.overview.setScrollAdjustment(adjustment, direction);
+
+                // Reset scroll on mapping
+                adjustment.value = 0;
             }));
     },
 
@@ -109,6 +112,13 @@ ViewByCategories.prototype = {
         this.actor.add(this._view.actor, { expand: true, x_fill: true, y_fill: true });
         this.actor.add(this._filters, { expand: false, y_fill: false, y_align: St.Align.START });
 
+        // Always select the "All" filter when switching to the app view
+        this.actor.connect('notify::mapped', Lang.bind(this,
+            function() {
+                if (this.actor.mapped && this._allFilter)
+                    this._selectCategory(-1);
+            }));
+
         this._sections = [];
     },
 
@@ -223,14 +233,20 @@ BaseAppSearchProvider.prototype = {
                  'icon': app.create_icon_texture(Search.RESULT_ICON_SIZE)};
     },
 
-    activateResult: function(id) {
+    activateResult: function(id, params) {
+        params = Params.parse(params, { workspace: null,
+                                        timestamp: null });
+
         let app = this._appSys.get_app(id);
-        app.activate();
+        app.activate(params.workspace ? params.workspace.index() : -1);
     },
 
-    dragActivateResult: function(id) {
+    dragActivateResult: function(id, params) {
+        params = Params.parse(params, { workspace: null,
+                                        timestamp: null });
+
         let app = this._appSys.get_app(id);
-        app.open_new_window();
+        app.open_new_window(params.workspace ? params.workspace.get_index() : -1);
     }
 };
 
@@ -319,15 +335,17 @@ function AppWellIcon(app, iconParams) {
 AppWellIcon.prototype = {
     _init : function(app, iconParams) {
         this.app = app;
-        this.actor = new St.Clickable({ style_class: 'app-well-app',
-                                         reactive: true,
-                                         x_fill: true,
-                                         y_fill: true });
+        this.actor = new St.Button({ style_class: 'app-well-app',
+                                     reactive: true,
+                                     button_mask: St.ButtonMask.ONE | St.ButtonMask.TWO,
+                                     x_fill: true,
+                                     y_fill: true });
         this.actor._delegate = this;
 
         this.icon = new AppIcon(app, iconParams);
         this.actor.set_child(this.icon.actor);
 
+        this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
         this.actor.connect('clicked', Lang.bind(this, this._onClicked));
 
         this._menu = null;
@@ -344,7 +362,6 @@ AppWellIcon.prototype = {
                Main.overview.endItemDrag(this);
             }));
 
-        this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
         this._menuTimeoutId = 0;
@@ -383,25 +400,25 @@ AppWellIcon.prototype = {
                 Lang.bind(this, function() {
                     this.popupMenu();
                 }));
-        }
-    },
-
-    _onClicked: function(actor, event) {
-        this._removeMenuTimeout();
-
-        let button = event.get_button();
-        if (button == 1) {
-            this._onActivate(event);
-        } else if (button == 2) {
-            let newWorkspace = Main.overview.workspaces.addWorkspace();
-            if (newWorkspace != null) {
-                newWorkspace.activate(global.get_current_time());
-                this.emit('launching');
-                this.app.open_new_window();
-                Main.overview.hide();
-            }
         } else if (button == 3) {
             this.popupMenu();
+            return true;
+        }
+        return false;
+    },
+
+    _onClicked: function(actor, button) {
+        this._removeMenuTimeout();
+
+        if (button == 1) {
+            this._onActivate(Clutter.get_current_event());
+        } else if (button == 2) {
+            // Last workspace is always empty
+            let launchWorkspace = global.screen.get_workspace_by_index(global.screen.n_workspaces - 1);
+            launchWorkspace.activate(global.get_current_time());
+            this.emit('launching');
+            this.app.open_new_window(-1);
+            Main.overview.hide();
         }
         return false;
     },
@@ -416,18 +433,12 @@ AppWellIcon.prototype = {
 
         if (!this._menu) {
             this._menu = new AppIconMenu(this);
-            this._menu.connect('highlight-window', Lang.bind(this, function (menu, window) {
-                this.highlightWindow(window);
-            }));
             this._menu.connect('activate-window', Lang.bind(this, function (menu, window) {
                 this.activateWindow(window);
             }));
             this._menu.connect('popup', Lang.bind(this, function (menu, isPoppedUp) {
-                if (isPoppedUp) {
-                    this._onMenuPoppedUp();
-                } else {
+                if (!isPoppedUp)
                     this._onMenuPoppedDown();
-                }
             }));
 
             this._menuManager.addMenu(this._menu);
@@ -438,45 +449,16 @@ AppWellIcon.prototype = {
         return false;
     },
 
-    highlightWindow: function(metaWindow) {
-        if (this._didActivateWindow)
-            return;
-        if (!this._getRunning())
-            return;
-        Main.overview.getWorkspacesForWindow(metaWindow).setHighlightWindow(metaWindow);
-    },
-
     activateWindow: function(metaWindow) {
         if (metaWindow) {
-            this._didActivateWindow = true;
             Main.activateWindow(metaWindow);
         } else {
             Main.overview.hide();
         }
     },
 
-    _onMenuPoppedUp: function() {
-        if (this._getRunning()) {
-            Main.overview.getWorkspacesForWindow(null).setApplicationWindowSelection(this.app.get_id());
-            this._setWindowSelection = true;
-            this._didActivateWindow = false;
-        }
-    },
-
     _onMenuPoppedDown: function() {
         this.actor.sync_hover();
-
-        if (this._didActivateWindow)
-            return;
-        if (!this._setWindowSelection)
-            return;
-
-        Main.overview.getWorkspacesForWindow(null).setApplicationWindowSelection(null);
-        this._setWindowSelection = false;
-    },
-
-    _getRunning: function() {
-        return this.app.state != Shell.AppState.STOPPED;
     },
 
     _onActivate: function (event) {
@@ -485,24 +467,22 @@ AppWellIcon.prototype = {
 
         if (modifiers & Clutter.ModifierType.CONTROL_MASK
             && this.app.state == Shell.AppState.RUNNING) {
-            this.app.open_new_window();
+            this.app.open_new_window(-1);
         } else {
-            this.app.activate();
+            this.app.activate(-1);
         }
         Main.overview.hide();
     },
 
-    // called by this._menuManager when it has the grab
-    menuEventFilter: function(event) {
-        return this._menu.menuEventFilter(event);
-    },
+    shellWorkspaceLaunch : function(params) {
+        params = Params.parse(params, { workspace: null,
+                                        timestamp: null });
 
-    shellWorkspaceLaunch : function() {
-        this.app.open_new_window();
+        this.app.open_new_window(params.workspace ? params.workspace.index() : -1);
     },
 
     getDragActor: function() {
-        return this.app.create_icon_texture(this.icon.iconSize);
+        return this.app.create_icon_texture(Main.overview.dash.iconSize);
     },
 
     // Returns the original actor that should align with the actor
@@ -525,11 +505,10 @@ AppIconMenu.prototype = {
         if (St.Widget.get_default_direction() == St.TextDirection.RTL)
             side = St.Side.RIGHT;
 
-        PopupMenu.PopupMenu.prototype._init.call(this, source.actor, St.Align.MIDDLE, side, 0);
+        PopupMenu.PopupMenu.prototype._init.call(this, source.actor, 0.5, side, 0);
 
         this._source = source;
 
-        this.connect('active-changed', Lang.bind(this, this._onActiveChanged));
         this.connect('activate', Lang.bind(this, this._onActivate));
         this.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
 
@@ -576,7 +555,6 @@ AppIconMenu.prototype = {
         this._toggleFavoriteMenuItem = this._appendMenuItem(isFavorite ? _("Remove from Favorites")
                                                                     : _("Add to Favorites"));
 
-        this._highlightedItem = null;
     },
 
     _appendSeparator: function () {
@@ -600,74 +578,16 @@ AppIconMenu.prototype = {
         if (open) {
             this.emit('popup', true);
         } else {
-            this._updateHighlight(null);
             this.emit('popup', false);
         }
     },
 
-    // called by this._menuManager when it has the grab
-    menuEventFilter: function(event) {
-        let eventType = event.type();
-
-        // Check if the user is interacting with a window representation
-        // rather than interacting with the menu
-
-        if (eventType == Clutter.EventType.BUTTON_RELEASE) {
-            let metaWindow = this._findMetaWindowForActor(event.get_source());
-            if (metaWindow)
-                this.emit('activate-window', metaWindow);
-        } else if (eventType == Clutter.EventType.ENTER) {
-            let metaWindow = this._findMetaWindowForActor(event.get_source());
-            if (metaWindow)
-                this._selectMenuItemForWindow(metaWindow, true);
-        } else if (eventType == Clutter.EventType.LEAVE) {
-            let metaWindow = this._findMetaWindowForActor(event.get_source());
-            if (metaWindow)
-                this._selectMenuItemForWindow(metaWindow, false);
-        }
-
-        return false;
-    },
-
-    _findMetaWindowForActor: function (actor) {
-        if (actor._delegate instanceof Workspace.WindowClone)
-            return actor._delegate.metaWindow;
-        else if (actor.get_meta_window)
-            return actor.get_meta_window();
-        return null;
-    },
-
-    _updateHighlight: function (item) {
-        if (this._highlightedItem)
-            this.emit('highlight-window', null);
-        this._highlightedItem = item;
-        if (this._highlightedItem) {
-            let window = this._highlightedItem._window;
-            if (window)
-                this.emit('highlight-window', window);
-        }
-    },
-
-    _selectMenuItemForWindow: function (metaWindow, selected) {
-        let items = this.getMenuItems();
-        for (let i = 0; i < items.length; i++) {
-            let item = items[i];
-            let menuMetaWindow = item._window;
-            if (menuMetaWindow == metaWindow)
-                item.setActive(selected);
-        }
-    },
-
-    _onActiveChanged: function (menu, child) {
-        this._updateHighlight(child);
-    },
-
     _onActivate: function (actor, child) {
         if (child._window) {
             let metaWindow = child._window;
             this.emit('activate-window', metaWindow);
         } else if (child == this._newWindowMenuItem) {
-            this._source.app.open_new_window();
+            this._source.app.open_new_window(-1);
             this.emit('activate-window', null);
         } else if (child == this._toggleFavoriteMenuItem) {
             let favs = AppFavorites.getAppFavorites();
