@@ -112,11 +112,11 @@ _Draggable.prototype = {
             return false;
 
         this._buttonDown = true;
-        // special case St.Clickable: grabbing the pointer would mess up the
+        // special case St.Button: grabbing the pointer would mess up the
         // internal state, so we start the drag manually on hover change
-        if (this.actor instanceof St.Clickable)
+        if (this.actor instanceof St.Button)
             this.actor.connect('notify::hover',
-                               Lang.bind(this, this._onClickableHoverChanged));
+                               Lang.bind(this, this._onButtonHoverChanged));
         else
             this._grabActor();
 
@@ -127,8 +127,8 @@ _Draggable.prototype = {
         return false;
     },
 
-    _onClickableHoverChanged: function(button) {
-        if (button.hover || !button.held)
+    _onButtonHoverChanged: function(button) {
+        if (button.hover || !button.pressed)
             return;
 
         button.fake_release();
@@ -461,12 +461,12 @@ _Draggable.prototype = {
             // its parent, adjusting for the fact that the parent
             // may have been moved or scaled
             let [parentX, parentY] = this._dragOrigParent.get_transformed_position();
-            x = parentX + this._dragOrigParent.scale_x * this._dragOrigX;
-            y = parentY + this._dragOrigParent.scale_y * this._dragOrigY;
-
             let [parentWidth, parentHeight] = this._dragOrigParent.get_size();
             let [parentScaledWidth, parentScaledHeight] = this._dragOrigParent.get_transformed_size();
             let parentScale = parentScaledWidth / parentWidth;
+
+            x = parentX + parentScale * this._dragOrigX;
+            y = parentY + parentScale * this._dragOrigY;
             scale = this._dragOrigScale * parentScale;
         } else {
             // Snap back actor to its original stage position
