@@ -137,12 +137,12 @@ ListItem.prototype = {
 
         let layout = new St.BoxLayout({ vertical: false});
 
-        this.actor = new St.Clickable({ style_class: 'end-session-dialog-app-list-item',
-                                        can_focus:   true,
-                                        child:       layout,
-                                        reactive:    true,
-                                        x_align:     St.Align.START,
-                                        x_fill:      true });
+        this.actor = new St.Button({ style_class: 'end-session-dialog-app-list-item',
+                                     can_focus:   true,
+                                     child:       layout,
+                                     reactive:    true,
+                                     x_align:     St.Align.START,
+                                     x_fill:      true });
 
         this._icon = this._app.create_icon_texture(_ITEM_ICON_SIZE);
 
@@ -171,7 +171,7 @@ ListItem.prototype = {
 
     _onClicked: function() {
         this.emit('activate');
-        this._app.activate();
+        this._app.activate(-1);
     }
 };
 Signals.addSignalMethods(ListItem.prototype);
@@ -470,7 +470,7 @@ EndSessionDialog.prototype = {
     OpenAsync: function(type, timestamp, totalSecondsToStayOpen, inhibitorObjectPaths, callback) {
         this._totalSecondsToStayOpen = totalSecondsToStayOpen;
         this._inhibitors = [];
-        this._applicationList.remove_all();
+        this._applicationList.destroy_children();
         this._type = type;
 
         if (!(this._type in DialogContent))
