@@ -11,16 +11,21 @@ const Pango = imports.gi.Pango;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const St = imports.gi.St;
-const Tweener = imports.ui.tweener;
 
-const Main = imports.ui.main;
 const BoxPointer = imports.ui.boxpointer;
+const GnomeSession = imports.misc.gnomeSession;
+const Main = imports.ui.main;
 const Params = imports.misc.params;
-const Utils = imports.misc.utils;
+const Tweener = imports.ui.tweener;
+const Util = imports.misc.util;
+
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const ANIMATION_TIME = 0.2;
 const NOTIFICATION_TIMEOUT = 4;
 const SUMMARY_TIMEOUT = 1;
+const LONGER_SUMMARY_TIMEOUT = 4;
 
 const HIDE_TIMEOUT = 0.2;
 const LONGER_HIDE_TIMEOUT = 0.6;
@@ -38,6 +43,17 @@ const State = {
     HIDING:  3
 };
 
+// These reasons are useful when we destroy the notifications received through
+// the notification daemon. We use EXPIRED for transient notifications that the
+// user did not interact with, DISMISSED for all other notifications that were
+// destroyed as a result of a user action, and SOURCE_CLOSED for the notifications
+// that were requested to be destroyed by the associated source.
+const NotificationDestroyedReason = {
+    EXPIRED: 1,
+    DISMISSED: 2,
+    SOURCE_CLOSED: 3
+};
+
 // Message tray has its custom Urgency enumeration. LOW, NORMAL and CRITICAL
 // urgency values map to the corresponding values for the notifications received
 // through the notification daemon. HIGH urgency value is used for chats received
@@ -75,8 +91,7 @@ URLHighlighter.prototype = {
         this.actor = new St.Label({ reactive: true, style_class: 'url-highlighter' });
         this._linkColor = '#ccccff';
         this.actor.connect('style-changed', Lang.bind(this, function() {
-            let color = new Clutter.Color();
-            let hasColor = this.actor.get_theme_node().get_color('link-color', color);
+            let [hasColor, color] = this.actor.get_theme_node().lookup_color('link-color', false);
             if (hasColor) {
                 let linkColor = color.to_string().substr(0, 7);
                 if (linkColor != this._linkColor) {
@@ -103,8 +118,7 @@ URLHighlighter.prototype = {
                     return true;
                 } catch (e) {
                     // TODO: remove this after gnome 3 release
-                    let p = new Shell.Process({ 'args' : ['gvfs-open', url] });
-                    p.run();
+                    Util.spawn(['gvfs-open', url]);
                     return true;
                 }
             }
@@ -135,13 +149,13 @@ URLHighlighter.prototype = {
 
         this.actor.clutter_text.set_markup(text);
         /* clutter_text.text contain text without markup */
-        this._urls = Utils.findUrls(this.actor.clutter_text.text);
+        this._urls = Util.findUrls(this.actor.clutter_text.text);
         this._highlightUrls();
     },
 
     _highlightUrls: function() {
         // text here contain markup
-        let urls = Utils.findUrls(this._text);
+        let urls = Util.findUrls(this._text);
         let markup = '';
         let pos = 0;
         for (let i = 0; i < urls.length; i++) {
@@ -175,6 +189,149 @@ URLHighlighter.prototype = {
     }
 };
 
+function FocusGrabber() {
+    this._init();
+}
+
+FocusGrabber.prototype = {
+    _init: function() {
+        this.actor = null;
+
+        this._hasFocus = false;
+        // We use this._prevFocusedWindow and this._prevKeyFocusActor to return the
+        // focus where it previously belonged after a focus grab, unless the user
+        // has explicitly changed that.
+        this._prevFocusedWindow = null;
+        this._prevKeyFocusActor = null;
+
+        this._focusActorChangedId = 0;
+        this._stageInputModeChangedId = 0;
+        this._capturedEventId = 0;
+        this._togglingFocusGrabMode = false;
+
+        Main.overview.connect('showing', Lang.bind(this,
+            function() {
+                this._toggleFocusGrabMode();
+            }));
+        Main.overview.connect('hidden', Lang.bind(this,
+            function() {
+                this._toggleFocusGrabMode();
+            }));
+    },
+
+    grabFocus: function(actor) {
+        if (this._hasFocus)
+            return;
+
+        this.actor = actor;
+
+        let metaDisplay = global.screen.get_display();
+
+        this._prevFocusedWindow = metaDisplay.focus_window;
+        this._prevKeyFocusActor = global.stage.get_key_focus();
+
+        if (!Main.overview.visible)
+            global.set_stage_input_mode(Shell.StageInputMode.FOCUSED);
+
+        // Use captured-event to notice clicks outside the focused actor
+        // without consuming them.
+        this._capturedEventId = global.stage.connect('captured-event', Lang.bind(this, this._onCapturedEvent));
+
+        this._stageInputModeChangedId = global.connect('notify::stage-input-mode', Lang.bind(this, this._stageInputModeChanged));
+        this._focusActorChangedId = global.stage.connect('notify::key-focus', Lang.bind(this, this._focusActorChanged));
+
+        this._hasFocus = true;
+
+        this.actor.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
+        this.emit('focus-grabbed');
+    },
+
+    _focusActorChanged: function() {
+        let focusedActor = global.stage.get_key_focus();
+        if (!focusedActor || !this.actor.contains(focusedActor)) {
+            this._prevKeyFocusActor = null;
+            this.ungrabFocus();
+        }
+    },
+
+    _stageInputModeChanged: function() {
+        this.ungrabFocus();
+    },
+
+    _onCapturedEvent: function(actor, event) {
+        let source = event.get_source();
+        switch (event.type()) {
+            case Clutter.EventType.BUTTON_PRESS:
+                if (!this.actor.contains(source))
+                    this.emit('button-pressed', source);
+                break;
+            case Clutter.EventType.KEY_PRESS:
+                let symbol = event.get_key_symbol();
+                if (symbol == Clutter.Escape) {
+                    this.emit('escape-pressed');
+                    return true;
+                }
+                break;
+        }
+
+        return false;
+    },
+
+    ungrabFocus: function() {
+        if (!this._hasFocus)
+            return;
+
+        let metaDisplay = global.screen.get_display();
+
+        if (this._focusActorChangedId > 0) {
+            global.stage.disconnect(this._focusActorChangedId);
+            this._focusActorChangedId = 0;
+        }
+
+        if (this._stageInputModeChangedId) {
+            global.disconnect(this._stageInputModeChangedId);
+            this._stageInputModeChangedId = 0;
+        }
+
+        if (this._capturedEventId > 0) {
+            global.stage.disconnect(this._capturedEventId);
+            this._capturedEventId = 0;
+        }
+
+        this._hasFocus = false;
+        this.emit('focus-ungrabbed');
+
+        if (this._prevFocusedWindow && !metaDisplay.focus_window) {
+            metaDisplay.set_input_focus_window(this._prevFocusedWindow, false, global.get_current_time());
+            this._prevFocusedWindow = null;
+        }
+        if (this._prevKeyFocusActor) {
+            global.stage.set_key_focus(this._prevKeyFocusActor);
+            this._prevKeyFocusActor = null;
+        } else {
+            // We don't want to keep any actor inside the previously focused actor focused.
+            let focusedActor = global.stage.get_key_focus();
+            if (focusedActor && this.actor.contains(focusedActor))
+                global.stage.set_key_focus(null);
+        }
+        if (!this._togglingFocusGrabMode)
+            this.actor = null;
+    },
+
+    // Because we grab focus differently in the overview
+    // and in the main view, we need to change how it is
+    // done when we move between the two.
+    _toggleFocusGrabMode: function() {
+        if (this._hasFocus) {
+            this._togglingFocusGrabMode = true;
+            this.ungrabFocus();
+            this.grabFocus(this.actor);
+            this._togglingFocusGrabMode = false;
+        }
+    }
+}
+Signals.addSignalMethods(FocusGrabber.prototype);
+
 // Notification:
 // @source: the notification's Source
 // @title: the title
@@ -241,6 +398,7 @@ Notification.prototype = {
         // 'transient' is a reserved keyword in JS, so we have to use an alternate variable name
         this.isTransient = false;
         this.expanded = false;
+        this._destroyed = false;
         this._useActionIcons = false;
         this._customContent = false;
         this._bannerBodyText = null;
@@ -248,20 +406,11 @@ Notification.prototype = {
         this._titleFitsInBannerMode = true;
         this._spacing = 0;
 
-        this._hasFocus = false;
-        this._lockTrayOnFocusGrab = false;
-        // We use this._prevFocusedWindow and this._prevKeyFocusActor to return the
-        // focus where it previously belonged after a focus grab, unless the user
-        // has explicitly changed that.
-        this._prevFocusedWindow = null;
-        this._prevKeyFocusActor = null;
-
-        this._focusActorChangedId = 0;
-        this._stageInputModeChangedId = 0;
-        this._capturedEventId = 0;
-        this._keyPressId = 0;
-
-        source.connect('destroy', Lang.bind(this, this.destroy));
+        source.connect('destroy', Lang.bind(this,
+            // Avoid passing 'source' as an argument to this.destroy()
+            function () {
+                this.destroy();
+            }));
 
         this.actor = new St.Table({ name: 'notification',
                                     reactive: true });
@@ -273,6 +422,8 @@ Notification.prototype = {
                     this._onClicked();
             }));
 
+        this._buttonFocusManager = St.FocusManager.get_for_stage(global.stage);
+
         // The first line should have the title, followed by the
         // banner text, but ellipsized if they won't both fit. We can't
         // make St.Table or St.BoxLayout do this the way we want (don't
@@ -294,15 +445,6 @@ Notification.prototype = {
         this._bannerBox.add_actor(this._bannerLabel);
 
         this.update(title, banner, params);
-
-        Main.overview.connect('showing', Lang.bind(this,
-            function() {
-                this._toggleFocusGrabMode();
-            }));
-        Main.overview.connect('hidden', Lang.bind(this,
-            function() {
-                this._toggleFocusGrabMode();
-            }));
     },
 
     // update:
@@ -384,7 +526,7 @@ Notification.prototype = {
             this._scrollArea = new St.ScrollView({ name: 'notification-scrollview',
                                                    vscrollbar_policy: Gtk.PolicyType.AUTOMATIC,
                                                    hscrollbar_policy: Gtk.PolicyType.NEVER,
-                                                   vshadows: true });
+                                                   vfade: true });
             this.actor.add(this._scrollArea, { row: 1,
                                                col: 1 });
             this._contentArea = new St.BoxLayout({ name: 'notification-body',
@@ -486,7 +628,7 @@ Notification.prototype = {
             this._buttonBox = box;
         }
 
-        let button = new St.Button();
+        let button = new St.Button({ can_focus: true });
 
         if (this._useActionIcons && Gtk.IconTheme.get_default().has_icon(id)) {
             button.add_style_class_name('notification-icon-button');
@@ -496,8 +638,13 @@ Notification.prototype = {
             button.label = label;
         }
 
+        if (this._buttonBox.get_children().length > 0)
+            this._buttonFocusManager.remove_group(this._buttonBox);
+
         this._buttonBox.add(button);
+        this._buttonFocusManager.add_group(this._buttonBox);
         button.connect('clicked', Lang.bind(this, this._onActionInvoked, id));
+
         this._updated();
     },
 
@@ -625,64 +772,7 @@ Notification.prototype = {
         this._bannerLabel.opacity = 255;
     },
 
-    grabFocus: function(lockTray) {
-        if (this._hasFocus)
-            return;
-
-        this._lockTrayOnFocusGrab = lockTray;
-
-        let metaDisplay = global.screen.get_display();
-
-        this._prevFocusedWindow = metaDisplay.focus_window;
-        this._prevKeyFocusActor = global.stage.get_key_focus();
-
-        if (!Main.overview.visible)
-            global.set_stage_input_mode(Shell.StageInputMode.FOCUSED);
-
-        // Use captured-event to notice clicks outside the notification
-        // without consuming them.
-        this._capturedEventId = global.stage.connect('captured-event', Lang.bind(this, this._onCapturedEvent));
-
-        this._stageInputModeChangedId = global.connect('notify::stage-input-mode', Lang.bind(this, this._stageInputModeChanged));
-        this._focusActorChangedId = global.stage.connect('notify::key-focus', Lang.bind(this, this._focusActorChanged));
-
-        this._hasFocus = true;
-        if (lockTray)
-            Main.messageTray.lock();
-    },
-
-    _focusActorChanged: function() {
-        let focusedActor = global.stage.get_key_focus();
-        if (!focusedActor || !this.actor.contains(focusedActor)) {
-            this._prevKeyFocusActor = null;
-            this.ungrabFocus();
-        }
-    },
-
-    _stageInputModeChanged: function() {
-        this.ungrabFocus();
-    },
-
-    _onCapturedEvent: function(actor, event) {
-        let source = event.get_source();
-        switch (event.type()) {
-            case Clutter.EventType.BUTTON_PRESS:
-                if (!this.actor.contains(source))
-                    this.ungrabFocus();
-                break;
-            case Clutter.EventType.KEY_PRESS:
-                let symbol = event.get_key_symbol();
-                if (symbol == Clutter.Escape) {
-                    Main.messageTray.escapeTray();
-                    return true;
-                }
-                break;
-        }
-
-        return false;
-    },
-
-    _onActionInvoked: function(actor, id) {
+    _onActionInvoked: function(actor, mouseButtonClicked, id) {
         this.emit('action-invoked', id);
         if (!this.resident) {
             // We don't hide a resident notification when the user invokes one of its actions,
@@ -704,66 +794,13 @@ Notification.prototype = {
             this.destroy();
     },
 
-    _onKeyPress: function(actor, event) {
-        let symbol = event.get_key_symbol();
-        if (symbol == Clutter.Escape) {
-            Main.messageTray.escapeTray();
-            return true;
-        }
-        return false;
-    },
-
-    ungrabFocus: function() {
-        if (!this._hasFocus)
+    destroy: function(reason) {
+        if (this._destroyed)
             return;
-
-        let metaDisplay = global.screen.get_display();
-
-        if (this._focusActorChangedId > 0) {
-            global.stage.disconnect(this._focusActorChangedId);
-            this._focusActorChangedId = 0;
-        }
-
-        if (this._stageInputModeChangedId) {
-            global.disconnect(this._stageInputModeChangedId);
-            this._stageInputModeChangedId = 0;
-        }
-
-        if (this._capturedEventId > 0) {
-            global.stage.disconnect(this._capturedEventId);
-            this._capturedEventId = 0;
-        }
-
-        this._hasFocus = false;
-        Main.messageTray.unlock();
-
-        if (this._prevFocusedWindow && !metaDisplay.focus_window) {
-            metaDisplay.set_input_focus_window(this._prevFocusedWindow, false, global.get_current_time());
-            this._prevFocusedWindow = null;
-        }
-        if (this._prevKeyFocusActor) {
-            global.stage.set_key_focus(this._prevKeyFocusActor);
-            this._prevKeyFocusActor = null;
-        } else {
-            // We don't want to keep the actor inside the notification focused.
-            let focusedActor = global.stage.get_key_focus();
-            if (focusedActor && this.actor.contains(focusedActor))
-                global.stage.set_key_focus(null);
-        }
-    },
-
-    // Because we grab focus differently in the overview
-    // and in the main view, we need to change how it is
-    // done when we move between the two.
-    _toggleFocusGrabMode: function() {
-        if (this._hasFocus) {
-            this.ungrabFocus();
-            this.grabFocus(this._lockTrayOnFocusGrab);
-        }
-    },
-
-    destroy: function() {
-        this.emit('destroy');
+        this._destroyed = true;
+        if (!reason)
+            reason = NotificationDestroyedReason.DISMISSED;
+        this.emit('destroy', reason);
     }
 };
 Signals.addSignalMethods(Notification.prototype);
@@ -780,6 +817,7 @@ Source.prototype = {
         this._iconBin = new St.Bin({ width: this.ICON_SIZE,
                                      height: this.ICON_SIZE });
         this.isTransient = false;
+        this.isChat = false;
     },
 
     setTransient: function(isTransient) {
@@ -904,6 +942,13 @@ function MessageTray() {
 
 MessageTray.prototype = {
     _init: function() {
+        this._presence = new GnomeSession.Presence();
+        this._userStatus = GnomeSession.PresenceStatus.AVAILABLE;
+        this._busy = false;
+        this._backFromAway = false;
+        this._presence.connect('StatusChanged', Lang.bind(this, this._onStatusChanged));
+        this._presence.getStatus(Lang.bind(this, this._onStatusChanged));
+
         this.actor = new St.Group({ name: 'message-tray',
                                     reactive: true,
                                     track_hover: true });
@@ -948,6 +993,21 @@ MessageTray.prototype = {
         // of the other items are collapsed.
         this._imaginarySummaryItemTitleWidth = 0;
 
+        this._focusGrabber = new FocusGrabber();
+        this._focusGrabber.connect('focus-grabbed', Lang.bind(this,
+            function() {
+                if (this._summaryNotification)
+                    this._lock();
+            }));
+        this._focusGrabber.connect('focus-ungrabbed', Lang.bind(this, this._unlock));
+        this._focusGrabber.connect('button-pressed', Lang.bind(this,
+           function(focusGrabber, source) {
+               if (this._clickedSummaryItem && !this._clickedSummaryItem.actor.contains(source))
+                   this._unsetClickedSummaryItem();
+               this._focusGrabber.ungrabFocus();
+           }));
+        this._focusGrabber.connect('escape-pressed', Lang.bind(this, this._escapeTray));
+
         this._trayState = State.HIDDEN;
         this._locked = false;
         this._useLongerTrayLeftTimeout = false;
@@ -978,18 +1038,22 @@ MessageTray.prototype = {
         Main.overview.connect('showing', Lang.bind(this,
             function() {
                 this._overviewVisible = true;
-                if (this._locked)
-                    this.unlock();
-                else
+                if (this._locked) {
+                    this._unsetClickedSummaryItem();
+                    this._unlock();
+                } else {
                     this._updateState();
+                }
             }));
         Main.overview.connect('hiding', Lang.bind(this,
             function() {
                 this._overviewVisible = false;
-                if (this._locked)
-                    this.unlock();
-                else
+                if (this._locked) {
+                    this._unsetClickedSummaryItem();
+                    this._unlock();
+                } else {
                     this._updateState();
+                }
             }));
 
         this._summaryItems = [];
@@ -999,6 +1063,7 @@ MessageTray.prototype = {
         // added to the summary without a notification being shown.
         this._newSummaryItems = [];
         this._longestSummaryItem = null;
+        this._chatSummaryItemsCount = 0;
     },
 
     _setSizePosition: function() {
@@ -1032,7 +1097,12 @@ MessageTray.prototype = {
 
         let summaryItem = new SummaryItem(source);
 
-        this._summary.insert_actor(summaryItem.actor, 0);
+        if (source.isChat) {
+            this._summary.insert_actor(summaryItem.actor, 0);
+            this._chatSummaryItemsCount++;
+        } else {
+            this._summary.insert_actor(summaryItem.actor, this._chatSummaryItemsCount);
+        }
 
         let titleWidth = summaryItem.getTitleNaturalWidth();
         if (titleWidth > this._summaryItemTitleWidth) {
@@ -1097,6 +1167,10 @@ MessageTray.prototype = {
             this._newSummaryItems.splice(newSummaryItemsIndex, 1);
 
         this._summaryItems.splice(index, 1);
+
+        if (source.isChat)
+            this._chatSummaryItemsCount--;
+
         if (this._longestSummaryItem.source == source) {
             let newTitleWidth = 0;
             this._longestSummaryItem = null;
@@ -1143,15 +1217,14 @@ MessageTray.prototype = {
             this._notificationQueue.splice(index, 1);
     },
 
-    lock: function() {
+    _lock: function() {
         this._locked = true;
     },
 
-    unlock: function() {
+    _unlock: function() {
         if (!this._locked)
             return;
         this._locked = false;
-        this._unsetClickedSummaryItem();
         this._updateState();
     },
 
@@ -1185,26 +1258,29 @@ MessageTray.prototype = {
     },
 
     _onSummaryItemHoverChanged: function(summaryItem) {
+        if (summaryItem.actor.hover)
+            this._setExpandedSummaryItem(summaryItem);
+    },
+
+    _setExpandedSummaryItem: function(summaryItem) {
+        if (summaryItem == this._expandedSummaryItem)
+            return;
+
         // We can't just animate individual summary items as the
         // pointer moves in and out of them, because if they don't
         // move in sync you get weird-looking wobbling. So whenever
         // there's a change, we have to re-tween the entire summary
         // area.
 
-        if (summaryItem.actor.hover) {
-            if (summaryItem == this._expandedSummaryItem)
-                return;
+        // Turn off ellipsization for the previously expanded item that is
+        // collapsing and for the item that is expanding because it looks
+        // better that way.
+        if (this._expandedSummaryItem)
+            this._expandedSummaryItem.setEllipsization(Pango.EllipsizeMode.NONE);
 
-            this._expandedSummaryItem = summaryItem;
-        } else {
-            if (summaryItem != this._expandedSummaryItem)
-                return;
-
-            this._expandedSummaryItem = null;
-
-            // Turn off ellipsization while collapsing; it looks better
-            summaryItem.setEllipsization(Pango.EllipsizeMode.NONE);
-        }
+        this._expandedSummaryItem = summaryItem;
+        if (this._expandedSummaryItem)
+            this._expandedSummaryItem.setEllipsization(Pango.EllipsizeMode.NONE);
 
         // We tween on a "_expandedSummaryItemTitleWidth" pseudo-property
         // that represents the current title width of the
@@ -1295,6 +1371,13 @@ MessageTray.prototype = {
             if (this._useLongerTrayLeftTimeout && !this._trayLeftTimeoutId)
                 return;
 
+            // Don't do anything if the mouse is over the summary notification as this should be considered as
+            // leaving the tray. The tray is locked when the summary notification is visible anyway, but we
+            // should treat the mouse being over the summary notification as the tray being left for collapsing
+            // any expanded summary item other than the one related to the notification.
+            if (this._summaryNotificationBoxPointer.bin.hover)
+                return;
+
             this._useLongerTrayLeftTimeout = false;
             if (this._trayLeftTimeoutId) {
                 Mainloop.source_remove(this._trayLeftTimeoutId);
@@ -1338,6 +1421,28 @@ MessageTray.prototype = {
         }
     },
 
+    _onStatusChanged: function(presence, status) {
+        this._backFromAway = (this._userStatus == GnomeSession.PresenceStatus.IDLE && this._userStatus != status);
+        this._userStatus = status;
+
+        if (status == GnomeSession.PresenceStatus.BUSY) {
+            // remove notification and allow the summary to be closed now
+            this._updateNotificationTimeout(0);
+            if (this._summaryTimeoutId) {
+                Mainloop.source_remove(this._summaryTimeoutId);
+                this._summaryTimeoutId = 0;
+            }
+            this._busy = true;
+        } else if (status != GnomeSession.PresenceStatus.IDLE) {
+            // We preserve the previous value of this._busy if the status turns to IDLE
+            // so that we don't start showing notifications queued during the BUSY state
+            // as the screensaver gets activated.
+            this._busy = false;
+        }
+
+        this._updateState();
+    },
+
     _onTrayLeftTimeout: function() {
         let [x, y, mods] = global.get_pointer();
         // We extend the timeout once if the mouse moved no further than MOUSE_LEFT_ACTOR_THRESHOLD to either side or up.
@@ -1361,8 +1466,8 @@ MessageTray.prototype = {
         return false;
     },
 
-    escapeTray: function() {
-        this.unlock();
+    _escapeTray: function() {
+        this._unlock();
         this._pointerInTray = false;
         this._pointerInSummary = false;
         this._updateNotificationTimeout(0);
@@ -1376,7 +1481,8 @@ MessageTray.prototype = {
     // at the present time.
     _updateState: function() {
         // Notifications
-        let notificationsPending = this._notificationQueue.length > 0;
+        let notificationsPending = this._notificationQueue.length > 0 &&
+                                   (!this._busy || this._notificationQueue[0].urgency == Urgency.CRITICAL);
         let notificationPinned = this._pointerInTray && !this._pointerInSummary && !this._notificationRemoved;
         let notificationExpanded = this._notificationBin.y < 0;
         let notificationExpired = (this._notificationTimeoutId == 0 && !(this._notification && this._notification.urgency == Urgency.CRITICAL) && !this._pointerInTray && !this._locked) || this._notificationRemoved;
@@ -1396,19 +1502,34 @@ MessageTray.prototype = {
         // Summary
         let summarySummoned = this._pointerInSummary || this._overviewVisible;
         let summaryPinned = this._summaryTimeoutId != 0 || this._pointerInTray || summarySummoned || this._locked;
+        let summaryHovered = this._pointerInTray || this._pointerInSummary;
+        let summaryVisibleWithNoHover = (this._overviewVisible || this._locked) && !summaryHovered;
+        let summaryNotificationIsForExpandedSummaryItem = (this._clickedSummaryItem == this._expandedSummaryItem);
 
         let notificationsVisible = (this._notificationState == State.SHOWING ||
                                     this._notificationState == State.SHOWN);
         let notificationsDone = !notificationsVisible && !notificationsPending;
 
         if (this._summaryState == State.HIDDEN) {
-            if (notificationsDone && this._newSummaryItems.length > 0)
-                this._showSummary(true);
-            else if (summarySummoned)
-                this._showSummary(false);
+            if (this._backFromAway) {
+                // Immediately set this to false, so that we don't schedule a timeout later
+                this._backFromAway = false;
+                if (!this._busy)
+                    this._showSummary(LONGER_SUMMARY_TIMEOUT);
+            } else if (notificationsDone && this._newSummaryItems.length > 0 && !this._busy) {
+                this._showSummary(SUMMARY_TIMEOUT);
+            } else if (summarySummoned) {
+                this._showSummary(0);
+            }
         } else if (this._summaryState == State.SHOWN) {
             if (!summaryPinned)
                 this._hideSummary();
+            else if (summaryVisibleWithNoHover && !summaryNotificationIsForExpandedSummaryItem)
+                // If we are hiding the summary, we'll collapse the expanded summary item when we are done
+                // so that there is no animation. However, we should collapse the expanded summary item
+                // if the summary is visible, but not hovered over, and the summary notification for the
+                // expanded summary item is not being shown.
+                this._setExpandedSummaryItem(null);
         }
 
         // Summary notification
@@ -1482,7 +1603,7 @@ MessageTray.prototype = {
     _showNotification: function() {
         this._notification = this._notificationQueue.shift();
         this._notificationClickedId = this._notification.connect('done-displaying',
-                                                                 Lang.bind(this, this.escapeTray));
+                                                                 Lang.bind(this, this._escapeTray));
         this._notificationBin.child = this._notification.actor;
 
         this._notificationBin.opacity = 0;
@@ -1507,23 +1628,40 @@ MessageTray.prototype = {
 
     _updateShowingNotification: function() {
         Tweener.removeTweens(this._notificationBin);
-        this._tween(this._notificationBin, '_notificationState', State.SHOWN,
-                    { y: 0,
-                      opacity: 255,
-                      time: ANIMATION_TIME,
-                      transition: 'easeOutQuad',
-                      onComplete: this._showNotificationCompleted,
-                      onCompleteScope: this
-                    });
 
         // We auto-expand notifications with CRITICAL urgency.
-        // We call _expandNotification() again on the notifications that
-        // are expanded in case they were in the process of hiding and need
-        // to re-expand.
+        // We use Tweener.removeTweens() to remove a tween that was hiding the notification we are
+        // updating, in case that notification was in the process of being hidden. However,
+        // Tweener.removeTweens() would also remove a tween that was updating the position of the
+        // notification we are updating, in case that notification was already expanded and its height
+        // changed. Therefore we need to call this._expandNotification() for expanded notifications
+        // to make sure their position is updated.
         if (this._notification.urgency == Urgency.CRITICAL || this._notification.expanded)
-            // This will overwrite the y tween, but leave the opacity
-            // tween, and so the onComplete will remain as well.
             this._expandNotification(true);
+
+        // We tween all notifications to full opacity. This ensures that both new notifications and
+        // notifications that might have been in the process of hiding get full opacity.
+        //
+        // We tween any notification showing in the banner mode to banner height (this._notificationBin.y = 0).
+        // This ensures that both new notifications and notifications in the banner mode that might
+        // have been in the process of hiding are shown with the banner height.
+        //
+        // We use this._showNotificationCompleted() onComplete callback to extend the time the updated
+        // notification is being shown.
+        //
+        // We don't set the y parameter for the tween for expanded notifications because
+        // this._expandNotification() will result in getting this._notificationBin.y set to the appropriate
+        // fully expanded value.
+        let tweenParams = { opacity: 255,
+                            time: ANIMATION_TIME,
+                            transition: 'easeOutQuad',
+                            onComplete: this._showNotificationCompleted,
+                            onCompleteScope: this
+                          };
+        if (!this._notification.expanded)
+            tweenParams.y = 0;
+
+        this._tween(this._notificationBin, '_notificationState', State.SHOWN, tweenParams);
    },
 
     _showNotificationCompleted: function() {
@@ -1560,7 +1698,7 @@ MessageTray.prototype = {
     },
 
     _hideNotification: function() {
-        this._notification.ungrabFocus();
+        this._focusGrabber.ungrabFocus();
         if (this._notificationExpandedId) {
             this._notification.disconnect(this._notificationExpandedId);
             this._notificationExpandedId = 0;
@@ -1586,13 +1724,13 @@ MessageTray.prototype = {
         let notification = this._notification;
         this._notification = null;
         if (notification.isTransient)
-            notification.destroy();
+            notification.destroy(NotificationDestroyedReason.EXPIRED);
     },
 
     _expandNotification: function(autoExpanding) {
         // Don't grab focus in notifications that are auto-expanded.
         if (!autoExpanding)
-            this._notification.grabFocus(false);
+            this._focusGrabber.grabFocus(this._notification.actor);
 
         if (!this._notificationExpandedId)
             this._notificationExpandedId =
@@ -1615,10 +1753,10 @@ MessageTray.prototype = {
     // We use this function to grab focus when the user moves the pointer
     // to a notification with CRITICAL urgency that was already auto-expanded.
     _ensureNotificationFocused: function() {
-        this._notification.grabFocus(false);
+        this._focusGrabber.grabFocus(this._notification.actor);
     },
 
-    _showSummary: function(withTimeout) {
+    _showSummary: function(timeout) {
         let primary = global.get_primary_monitor();
         this._summaryBin.opacity = 0;
         this._summaryBin.y = this.actor.height;
@@ -1629,16 +1767,16 @@ MessageTray.prototype = {
                       transition: 'easeOutQuad',
                       onComplete: this._showSummaryCompleted,
                       onCompleteScope: this,
-                      onCompleteParams: [withTimeout]
+                      onCompleteParams: [timeout]
                     });
     },
 
-    _showSummaryCompleted: function(withTimeout) {
+    _showSummaryCompleted: function(timeout) {
         this._newSummaryItems = [];
 
-        if (withTimeout) {
+        if (timeout != 0) {
             this._summaryTimeoutId =
-                Mainloop.timeout_add(SUMMARY_TIMEOUT * 1000,
+                Mainloop.timeout_add(timeout * 1000,
                                      Lang.bind(this, this._summaryTimeout));
         }
     },
@@ -1653,21 +1791,28 @@ MessageTray.prototype = {
         this._tween(this._summaryBin, '_summaryState', State.HIDDEN,
                     { opacity: 0,
                       time: ANIMATION_TIME,
-                      transition: 'easeOutQuad'
+                      transition: 'easeOutQuad',
+                      onComplete: this._hideSummaryCompleted,
+                      onCompleteScope: this,
                     });
         this._newSummaryItems = [];
     },
 
+    _hideSummaryCompleted: function() {
+        this._expandedSummaryItem = null;
+        this._expandedSummaryItemTitleWidth = this._summaryItemTitleWidth;
+    },
+
     _showSummaryNotification: function() {
         this._summaryNotification = this._clickedSummaryItem.source.notification;
         this._summaryNotificationClickedId = this._summaryNotification.connect('done-displaying',
-                                                                               Lang.bind(this, this.escapeTray));
+                                                                               Lang.bind(this, this._escapeTray));
         let index = this._notificationQueue.indexOf(this._summaryNotification);
         if (index != -1)
             this._notificationQueue.splice(index, 1);
 
         this._summaryNotificationBoxPointer.bin.child = this._summaryNotification.actor;
-        this._summaryNotification.grabFocus(true);
+        this._focusGrabber.grabFocus(this._summaryNotification.actor);
 
         if (!this._summaryNotificationExpandedId)
             this._summaryNotificationExpandedId = this._summaryNotification.connect('expanded', Lang.bind(this, this._onSummaryNotificationExpanded));
@@ -1695,7 +1840,7 @@ MessageTray.prototype = {
         if (!this._clickedSummaryItem)
             return;
 
-        this._summaryNotificationBoxPointer.setPosition(this._clickedSummaryItem.actor, 0, St.Align.MIDDLE);
+        this._summaryNotificationBoxPointer.setPosition(this._clickedSummaryItem.actor, 0, 0.5);
     },
 
     _unsetClickedSummaryItem: function() {
@@ -1722,7 +1867,7 @@ MessageTray.prototype = {
         if (this._summaryState != State.SHOWN)
             this._unsetClickedSummaryItem();
 
-        this._summaryNotification.ungrabFocus();
+        this._focusGrabber.ungrabFocus();
         this._summaryNotificationState = State.HIDING;
         this._summaryNotificationBoxPointer.hide(true, Lang.bind(this, this._hideSummaryNotificationCompleted));
     },
@@ -1736,10 +1881,36 @@ MessageTray.prototype = {
         let summaryNotification = this._summaryNotification;
         this._summaryNotification = null;
         if (summaryNotification.isTransient && !this._reNotifyWithSummaryNotificationAfterHide)
-            summaryNotification.destroy();
+            summaryNotification.destroy(NotificationDestroyedReason.EXPIRED);
         if (this._reNotifyWithSummaryNotificationAfterHide) {
             this._onNotify(summaryNotification.source, summaryNotification);
             this._reNotifyWithSummaryNotificationAfterHide = false;
         }
+        if (this._clickedSummaryItem)
+            this._updateState();
+    }
+};
+
+function SystemNotificationSource() {
+    this._init();
+}
+
+SystemNotificationSource.prototype = {
+    __proto__:  Source.prototype,
+
+    _init: function() {
+        Source.prototype._init.call(this, _("System Information"));
+
+        this._setSummaryIcon(this.createNotificationIcon());
+    },
+
+    createNotificationIcon: function() {
+        return new St.Icon({ icon_name: 'dialog-information',
+                             icon_type: St.IconType.SYMBOLIC,
+                             icon_size: this.ICON_SIZE });
+    },
+
+    _notificationClicked: function() {
+        this.destroy();
     }
 };
