@@ -1,6 +1,7 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const Clutter = imports.gi.Clutter;
+const Gtk = imports.gi.Gtk;
 const Meta = imports.gi.Meta;
 const Mainloop = imports.mainloop;
 const Signals = imports.signals;
@@ -15,7 +16,6 @@ const AppDisplay = imports.ui.appDisplay;
 const Dash = imports.ui.dash;
 const DND = imports.ui.dnd;
 const DocDisplay = imports.ui.docDisplay;
-const GenericDisplay = imports.ui.genericDisplay;
 const Lightbox = imports.ui.lightbox;
 const Main = imports.ui.main;
 const MessageTray = imports.ui.messageTray;
@@ -31,32 +31,19 @@ const ANIMATION_TIME = 0.25;
 // We split the screen vertically between the dash and the view selector.
 const DASH_SPLIT_FRACTION = 0.1;
 
-
 const DND_WINDOW_SWITCH_TIMEOUT = 1250;
 
-function Source() {
-    this._init();
-}
+const SwipeScrollDirection = {
+    NONE: 0,
+    HORIZONTAL: 1,
+    VERTICAL: 2
+};
 
-Source.prototype = {
-    __proto__:  MessageTray.Source.prototype,
-
-    _init: function() {
-        MessageTray.Source.prototype._init.call(this,
-                                                "System Information");
-        this._setSummaryIcon(this.createNotificationIcon());
-    },
-
-    createNotificationIcon: function() {
-        return new St.Icon({ icon_name: 'dialog-information',
-                             icon_type: St.IconType.SYMBOLIC,
-                             icon_size: this.ICON_SIZE });
-    },
-
-    _notificationClicked: function() {
-        this.destroy();
-    }
-}
+const SwipeScrollResult = {
+    CANCEL: 0,
+    SWIPE: 1,
+    CLICK: 2
+};
 
 function ShellInfo() {
     this._init();
@@ -79,7 +66,7 @@ ShellInfo.prototype = {
 
     setMessage: function(text, undoCallback, undoLabel) {
         if (this._source == null) {
-            this._source = new Source();
+            this._source = new MessageTray.SystemNotificationSource();
             this._source.connect('destroy', Lang.bind(this,
                 function() {
                     this._source = null;
@@ -124,7 +111,8 @@ Overview.prototype = {
 
         this._spacing = 0;
 
-        this._group = new St.Group({ name: 'overview' });
+        this._group = new St.Group({ name: 'overview',
+                                     reactive: true });
         this._group._delegate = this;
         this._group.connect('style-changed',
             Lang.bind(this, function() {
@@ -136,7 +124,10 @@ Overview.prototype = {
                 }
             }));
 
-        this.shellInfo = new ShellInfo();
+        this._scrollDirection = SwipeScrollDirection.NONE;
+        this._scrollAdjustment = null;
+        this._capturedEventId = 0;
+        this._buttonPressId = 0;
 
         this._workspacesDisplay = null;
 
@@ -159,6 +150,32 @@ Overview.prototype = {
         this._group.hide();
         global.overlay_group.add_actor(this._group);
 
+        this._coverPane.hide();
+
+        // XDND
+        this._dragMonitor = {
+            dragMotion: Lang.bind(this, this._onDragMotion)
+        };
+
+        Main.xdndHandler.connect('drag-begin', Lang.bind(this, this._onDragBegin));
+        Main.xdndHandler.connect('drag-end', Lang.bind(this, this._onDragEnd));
+
+        this._windowSwitchTimeoutId = 0;
+        this._windowSwitchTimestamp = 0;
+        this._lastActiveWorkspaceIndex = -1;
+        this._lastHoveredWindow = null;
+        this._needsFakePointerEvent = false;
+
+        this.workspaces = null;
+    },
+
+    // The members we construct that are implemented in JS might
+    // want to access the overview as Main.overview to connect
+    // signal handlers and so forth. So we create them after
+    // construction in this init() method.
+    init: function() {
+        this.shellInfo = new ShellInfo();
+
         this.viewSelector = new ViewSelector.ViewSelector();
         this._group.add_actor(this.viewSelector.actor);
 
@@ -175,27 +192,11 @@ Overview.prototype = {
         this.viewSelector.addSearchProvider(new DocDisplay.DocSearchProvider());
 
         // TODO - recalculate everything when desktop size changes
-        this._dash = new Dash.Dash();
-        this._group.add_actor(this._dash.actor);
-        this._dash.actor.add_constraint(this.viewSelector.constrainY);
-        this._dash.actor.add_constraint(this.viewSelector.constrainHeight);
+        this.dash = new Dash.Dash();
+        this._group.add_actor(this.dash.actor);
+        this.dash.actor.add_constraint(this.viewSelector.constrainY);
+        this.dash.actor.add_constraint(this.viewSelector.constrainHeight);
 
-        this._coverPane.lower_bottom();
-
-        // XDND
-        this._dragMonitor = {
-            dragMotion: Lang.bind(this, this._onDragMotion)
-        };
-
-        Main.xdndHandler.connect('drag-begin', Lang.bind(this, this._onDragBegin));
-        Main.xdndHandler.connect('drag-end', Lang.bind(this, this._onDragEnd));
-
-        this._windowSwitchTimeoutId = 0;
-        this._windowSwitchTimestamp = 0;
-        this._lastActiveWorkspaceIndex = -1;
-        this._needsFakePointerEvent = false;
-
-        this.workspaces = null;
     },
 
     _onDragBegin: function() {
@@ -212,7 +213,7 @@ Overview.prototype = {
             global.screen.get_workspace_by_index(this._lastActiveWorkspaceIndex).activate(time);
             this.hideTemporarily();
         }
-
+        this._lastHoveredWindow = null;
         DND.removeMonitor(this._dragMonitor);
     },
 
@@ -226,15 +227,24 @@ Overview.prototype = {
     },
 
     _onDragMotion: function(dragEvent) {
+        let targetIsWindow = dragEvent.targetActor &&
+                             dragEvent.targetActor._delegate &&
+                             dragEvent.targetActor._delegate.metaWindow;
+
+        if (targetIsWindow &&
+            dragEvent.targetActor._delegate.metaWindow == this._lastHoveredWindow)
+            return DND.DragMotionResult.CONTINUE;
+
+        this._lastHoveredWindow = null;
+
         if (this._windowSwitchTimeoutId != 0) {
             Mainloop.source_remove(this._windowSwitchTimeoutId);
             this._windowSwitchTimeoutId = 0;
             this._needsFakePointerEvent = false;
         }
 
-        if (dragEvent.targetActor &&
-            dragEvent.targetActor._delegate &&
-            dragEvent.targetActor._delegate.metaWindow) {
+        if (targetIsWindow) {
+            this._lastHoveredWindow = dragEvent.targetActor._delegate.metaWindow;
             this._windowSwitchTimestamp = global.get_current_time();
             this._windowSwitchTimeoutId = Mainloop.timeout_add(DND_WINDOW_SWITCH_TIMEOUT,
                                             Lang.bind(this, function() {
@@ -242,12 +252,164 @@ Overview.prototype = {
                                                 Main.activateWindow(dragEvent.targetActor._delegate.metaWindow,
                                                                     this._windowSwitchTimestamp);
                                                 this.hideTemporarily();
+                                                this._lastHoveredWindow = null;
                                             }));
         }
 
         return DND.DragMotionResult.CONTINUE;
     },
 
+    setScrollAdjustment: function(adjustment, direction) {
+        this._scrollAdjustment = adjustment;
+        if (this._scrollAdjustment == null)
+            this._scrollDirection = SwipeScrollDirection.NONE;
+        else
+            this._scrollDirection = direction;
+    },
+
+    _onButtonPress: function(actor, event) {
+        if (this._scrollDirection == SwipeScrollDirection.NONE)
+            return;
+
+        let [stageX, stageY] = event.get_coords();
+        this._dragStartX = this._dragX = stageX;
+        this._dragStartY = this._dragY = stageY;
+        this._dragStartValue = this._scrollAdjustment.value;
+        this._lastMotionTime = -1; // used to track "stopping" while swipe-scrolling
+        this._capturedEventId = global.stage.connect('captured-event',
+            Lang.bind(this, this._onCapturedEvent));
+        this.emit('swipe-scroll-begin');
+    },
+
+    _onCapturedEvent: function(actor, event) {
+        let stageX, stageY;
+        let threshold = Gtk.Settings.get_default().gtk_dnd_drag_threshold;
+
+        switch(event.type()) {
+            case Clutter.EventType.BUTTON_RELEASE:
+                [stageX, stageY] = event.get_coords();
+
+                // default to snapping back to the original value
+                let newValue = this._dragStartValue;
+
+                let minValue = this._scrollAdjustment.lower;
+                let maxValue = this._scrollAdjustment.upper - this._scrollAdjustment.page_size;
+
+                let direction;
+                if (this._scrollDirection == SwipeScrollDirection.HORIZONTAL) {
+                    direction = stageX > this._dragStartX ? -1 : 1;
+                    if (St.Widget.get_default_direction() == St.TextDirection.RTL)
+                        direction *= -1;
+                } else {
+                    direction = stageY > this._dragStartY ? -1 : 1;
+                }
+
+                // We default to scroll a full page size; both the first
+                // and the last page may be smaller though, so we need to
+                // adjust difference in those cases.
+                let difference = direction * this._scrollAdjustment.page_size;
+                if (this._dragStartValue + difference > maxValue)
+                    difference = maxValue - this._dragStartValue;
+                else if (this._dragStartValue + difference < minValue)
+                    difference = minValue - this._dragStartValue;
+
+                // If the user has moved more than half the scroll
+                // difference, we want to "settle" to the new value
+                // even if the user stops dragging rather "throws" by
+                // releasing during the drag.
+                let distance = this._dragStartValue - this._scrollAdjustment.value;
+                let noStop = Math.abs(distance / difference) > 0.5;
+
+                // We detect if the user is stopped by comparing the
+                // timestamp of the button release with the timestamp of
+                // the last motion. Experimentally, a difference of 0 or 1
+                // millisecond indicates that the mouse is in motion, a
+                // larger difference indicates that the mouse is stopped.
+                if ((this._lastMotionTime > 0 &&
+                     this._lastMotionTime > event.get_time() - 2) ||
+                    noStop) {
+                    if (this._dragStartValue + difference >= minValue &&
+                        this._dragStartValue + difference <= maxValue)
+                        newValue += difference;
+                }
+
+                let result;
+
+                // See if the user has moved the mouse enough to trigger
+                // a drag
+                if (Math.abs(stageX - this._dragStartX) < threshold &&
+                    Math.abs(stageY - this._dragStartY) < threshold) {
+                    // no motion? It's a click!
+                    result = SwipeScrollResult.CLICK;
+                    this.emit('swipe-scroll-end', result);
+                } else {
+                    if (newValue == this._dragStartValue)
+                        result = SwipeScrollResult.CANCEL;
+                    else
+                        result = SwipeScrollResult.SWIPE;
+
+                    // The event capture handler is disconnected
+                    // while scrolling to the final position, so
+                    // to avoid undesired prelights we raise
+                    // the cover pane.
+                    this._coverPane.raise_top();
+                    this._coverPane.show();
+
+                    Tweener.addTween(this._scrollAdjustment,
+                                     { value: newValue,
+                                       time: ANIMATION_TIME,
+                                       transition: 'easeOutQuad',
+                                       onCompleteScope: this,
+                                       onComplete: function() {
+                                          this._coverPane.hide();
+                                          this.emit('swipe-scroll-end',
+                                                    result);
+                                       }
+                                     });
+                }
+
+                global.stage.disconnect(this._capturedEventId);
+                this._capturedEventId = 0;
+
+                return result != SwipeScrollResult.CLICK;
+
+            case Clutter.EventType.MOTION:
+                [stageX, stageY] = event.get_coords();
+                let dx = this._dragX - stageX;
+                let dy = this._dragY - stageY;
+                let primary = global.get_primary_monitor();
+
+                this._dragX = stageX;
+                this._dragY = stageY;
+                this._lastMotionTime = event.get_time();
+
+                // See if the user has moved the mouse enough to trigger
+                // a drag
+                if (Math.abs(stageX - this._dragStartX) < threshold &&
+                    Math.abs(stageY - this._dragStartY) < threshold)
+                    return true;
+
+                if (this._scrollDirection == SwipeScrollDirection.HORIZONTAL) {
+                    if (St.Widget.get_default_direction() == St.TextDirection.RTL)
+                        this._scrollAdjustment.value -= (dx / primary.width) * this._scrollAdjustment.page_size;
+                    else
+                        this._scrollAdjustment.value += (dx / primary.width) * this._scrollAdjustment.page_size;
+                } else {
+                    this._scrollAdjustment.value += (dy / primary.height) * this._scrollAdjustment.page_size;
+                }
+
+                return true;
+
+            // Block enter/leave events to avoid prelights
+            // during swipe-scroll
+            case Clutter.EventType.ENTER:
+            case Clutter.EventType.LEAVE:
+                return true;
+        }
+
+        return false;
+    },
+
     _getDesktopClone: function() {
         let windows = global.get_window_actors().filter(function(w) {
             return w.meta_window.get_window_type() == Meta.WindowType.DESKTOP;
@@ -275,21 +437,21 @@ Overview.prototype = {
         this._coverPane.set_position(0, contentY);
         this._coverPane.set_size(primary.width, contentHeight);
 
-        let viewWidth = (1.0 - DASH_SPLIT_FRACTION) * primary.width - this._spacing;
+        let dashWidth = Math.round(DASH_SPLIT_FRACTION * primary.width);
+        let viewWidth = primary.width - dashWidth - this._spacing;
         let viewHeight = contentHeight - 2 * this._spacing;
         let viewY = contentY + this._spacing;
-        let viewX = rtl ? 0
-                        : Math.floor(DASH_SPLIT_FRACTION * primary.width) + this._spacing;
+        let viewX = rtl ? 0 : dashWidth + this._spacing;
 
         // Set the dash's x position - y is handled by a constraint
         let dashX;
         if (rtl) {
-            this._dash.actor.set_anchor_point_from_gravity(Clutter.Gravity.NORTH_EAST);
+            this.dash.actor.set_anchor_point_from_gravity(Clutter.Gravity.NORTH_EAST);
             dashX = primary.width;
         } else {
             dashX = 0;
         }
-        this._dash.actor.set_x(dashX);
+        this.dash.actor.set_x(dashX);
 
         this.viewSelector.actor.set_position(viewX, viewY);
         this.viewSelector.actor.set_size(viewWidth, viewHeight);
@@ -350,6 +512,9 @@ Overview.prototype = {
         this._modal = true;
         this._animateVisible();
         this._shown = true;
+
+        this._buttonPressId = this._group.connect('button-press-event',
+            Lang.bind(this, this._onButtonPress));
     },
 
     _animateVisible: function() {
@@ -370,9 +535,7 @@ Overview.prototype = {
         this._group.show();
         this._background.show();
 
-        this.viewSelector.show();
         this._workspacesDisplay.show();
-        this._dash.show();
 
         this.workspaces = this._workspacesDisplay.workspacesView;
         global.overlay_group.add_actor(this.workspaces.actor);
@@ -417,6 +580,7 @@ Overview.prototype = {
                          });
 
         this._coverPane.raise_top();
+        this._coverPane.show();
         this.emit('showing');
     },
 
@@ -447,6 +611,10 @@ Overview.prototype = {
 
         this._shown = false;
         this._syncInputMode();
+
+        if (this._buttonPressId > 0)
+            this._group.disconnect(this._buttonPressId);
+        this._buttonPressId = 0;
     },
 
     // hideTemporarily:
@@ -493,20 +661,20 @@ Overview.prototype = {
 
         if (this._shown) {
             if (!this._modal) {
-                if (Main.pushModal(this._dash.actor))
+                if (Main.pushModal(this.dash.actor))
                     this._modal = true;
                 else
                     this.hide();
             }
         } else if (this._shownTemporarily) {
             if (this._modal) {
-                Main.popModal(this._dash.actor);
+                Main.popModal(this.dash.actor);
                 this._modal = false;
             }
             global.stage_input_mode = Shell.StageInputMode.FULLSCREEN;
         } else {
             if (this._modal) {
-                Main.popModal(this._dash.actor);
+                Main.popModal(this.dash.actor);
                 this._modal = false;
             }
             else if (global.stage_input_mode == Shell.StageInputMode.FULLSCREEN)
@@ -556,13 +724,14 @@ Overview.prototype = {
                          });
 
         this._coverPane.raise_top();
+        this._coverPane.show();
         this.emit('hiding');
     },
 
     _showDone: function() {
         this.animationInProgress = false;
         this._desktopFade.hide();
-        this._coverPane.lower_bottom();
+        this._coverPane.hide();
 
         this.emit('shown');
         // Handle any calls to hide* while we were showing
@@ -579,8 +748,6 @@ Overview.prototype = {
         this.workspaces = null;
 
         this._workspacesDisplay.hide();
-        this.viewSelector.hide();
-        this._dash.hide();
 
         this._desktopFade.hide();
         this._background.hide();
@@ -590,7 +757,7 @@ Overview.prototype = {
         this.animationInProgress = false;
         this._hideInProgress = false;
 
-        this._coverPane.lower_bottom();
+        this._coverPane.hide();
 
         this.emit('hidden');
         // Handle any calls to show* while we were hiding
