@@ -18,35 +18,6 @@ const _ = Gettext.gettext;
 
 const SLIDER_SCROLL_STEP = 0.05; /* Slider scrolling step in % */
 
-function Switch() {
-    this._init.apply(this, arguments);
-}
-
-Switch.prototype = {
-    _init: function(state) {
-        this.actor = new St.Bin({ style_class: 'toggle-switch' });
-        // Translators: this MUST be either "toggle-switch-us"
-        // (for toggle switches containing the English words
-        // "ON" and "OFF") or "toggle-switch-intl" (for toggle
-        // switches containing "◯" and "|"). Other values will
-        // simply result in invisible toggle switches.
-        this.actor.add_style_class_name(_("toggle-switch-us"));
-        this.setToggleState(state);
-    },
-
-    setToggleState: function(state) {
-        if (state)
-            this.actor.add_style_pseudo_class('checked');
-        else
-            this.actor.remove_style_pseudo_class('checked');
-        this.state = state;
-    },
-
-    toggle: function() {
-        this.setToggleState(!this.state);
-    }
-};
-
 function PopupBaseMenuItem(params) {
     this._init(params);
 }
@@ -55,7 +26,9 @@ PopupBaseMenuItem.prototype = {
     _init: function (params) {
         params = Params.parse (params, { reactive: true,
                                          activate: true,
-                                         hover: true });
+                                         hover: true,
+                                         style_class: null
+                                       });
         this.actor = new Shell.GenericContainer({ style_class: 'popup-menu-item',
                                                   reactive: params.reactive,
                                                   track_hover: params.reactive,
@@ -72,6 +45,9 @@ PopupBaseMenuItem.prototype = {
         this._spacing = 0;
         this.active = false;
 
+        if (params.style_class)
+            this.actor.add_style_class_name(params.style_class);
+
         if (params.reactive && params.activate) {
             this.actor.connect('button-release-event', Lang.bind(this, this._onButtonReleaseEvent));
             this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
@@ -180,8 +156,7 @@ PopupBaseMenuItem.prototype = {
     _onRepaintDot: function(area) {
         let cr = area.get_context();
         let [width, height] = area.get_surface_size();
-        let color = new Clutter.Color();
-        area.get_theme_node().get_foreground_color(color);
+        let color = area.get_theme_node().get_foreground_color();
 
         cr.setSourceRGBA (
             color.red / 255,
@@ -338,10 +313,8 @@ PopupSeparatorMenuItem.prototype = {
         let [width, height] = area.get_surface_size();
         let margin = themeNode.get_length('-margin-horizontal');
         let gradientHeight = themeNode.get_length('-gradient-height');
-        let startColor = new Clutter.Color();
-        themeNode.get_color('-gradient-start', startColor);
-        let endColor = new Clutter.Color();
-        themeNode.get_color('-gradient-end', endColor);
+        let startColor = themeNode.get_color('-gradient-start');
+        let endColor = themeNode.get_color('-gradient-end');
 
         let gradientWidth = (width - margin * 2);
         let gradientOffset = (height - gradientHeight) / 2;
@@ -355,6 +328,107 @@ PopupSeparatorMenuItem.prototype = {
     }
 };
 
+const PopupAlternatingMenuItemState = {
+    DEFAULT: 0,
+    ALTERNATIVE: 1
+}
+
+function PopupAlternatingMenuItem() {
+    this._init.apply(this, arguments);
+}
+
+PopupAlternatingMenuItem.prototype = {
+    __proto__: PopupBaseMenuItem.prototype,
+
+    _init: function(text, alternateText, params) {
+        PopupBaseMenuItem.prototype._init.call(this, params);
+        this.actor.add_style_class_name('popup-alternating-menu-item');
+
+        this._text = text;
+        this._alternateText = alternateText;
+        this.label = new St.Label({ text: text });
+        this.state = PopupAlternatingMenuItemState.DEFAULT;
+        this.addActor(this.label);
+
+        this.actor.connect('notify::mapped', Lang.bind(this, this._onMapped));
+    },
+
+    _onMapped: function() {
+        if (this.actor.mapped) {
+            this._capturedEventId = global.stage.connect('captured-event',
+                                                         Lang.bind(this, this._onCapturedEvent));
+            this._updateStateFromModifiers();
+        } else {
+            if (this._capturedEventId != 0) {
+                global.stage.disconnect(this._capturedEventId);
+                this._capturedEventId = 0;
+            }
+        }
+    },
+
+    _setState: function(state) {
+        if (this.state != state) {
+            if (state == PopupAlternatingMenuItemState.ALTERNATIVE && !this._canAlternate())
+                return;
+
+            this.state = state;
+            this._updateLabel();
+        }
+    },
+
+    _updateStateFromModifiers: function() {
+        let [x, y, mods] = global.get_pointer();
+        let state;
+
+        if ((mods & Clutter.ModifierType.MOD1_MASK) == 0) {
+            state = PopupAlternatingMenuItemState.DEFAULT;
+        } else {
+            state = PopupAlternatingMenuItemState.ALTERNATIVE;
+        }
+
+        this._setState(state);
+    },
+
+    _onCapturedEvent: function(actor, event) {
+        if (event.type() != Clutter.EventType.KEY_PRESS &&
+            event.type() != Clutter.EventType.KEY_RELEASE)
+            return false;
+
+        let key = event.get_key_symbol();
+
+        if (key == Clutter.KEY_Alt_L || key == Clutter.KEY_Alt_R)
+            this._updateStateFromModifiers();
+
+        return false;
+    },
+
+    _updateLabel: function() {
+        if (this.state == PopupAlternatingMenuItemState.ALTERNATIVE) {
+            this.actor.add_style_pseudo_class('alternate');
+            this.label.set_text(this._alternateText);
+        } else {
+            this.actor.remove_style_pseudo_class('alternate');
+            this.label.set_text(this._text);
+        }
+    },
+
+    _canAlternate: function() {
+        if (this.state == PopupAlternatingMenuItemState.DEFAULT && !this._alternateText)
+            return false;
+        return true;
+    },
+
+    updateText: function(text, alternateText) {
+        this._text = text;
+        this._alternateText = alternateText;
+
+        if (!this._canAlternate())
+            this._setState(PopupAlternatingMenuItemState.DEFAULT);
+
+        this._updateLabel();
+    }
+};
+
 function PopupSliderMenuItem() {
     this._init.apply(this, arguments);
 }
@@ -402,10 +476,8 @@ PopupSliderMenuItem.prototype = {
 
         let sliderBorderWidth = themeNode.get_length('-slider-border-width');
 
-        let sliderBorderColor = new Clutter.Color();
-        themeNode.get_color('-slider-border-color', sliderBorderColor);
-        let sliderColor = new Clutter.Color();
-        themeNode.get_color('-slider-background-color', sliderColor);
+        let sliderBorderColor = themeNode.get_color('-slider-border-color');
+        let sliderColor = themeNode.get_color('-slider-background-color');
 
         cr.setSourceRGBA (
             sliderColor.red / 255,
@@ -425,8 +497,7 @@ PopupSliderMenuItem.prototype = {
         let handleY = height / 2;
         let handleX = handleRadius + (width - 2 * handleRadius) * this._value;
 
-        let color = new Clutter.Color();
-        themeNode.get_foreground_color(color);
+        let color = themeNode.get_foreground_color();
         cr.setSourceRGBA (
             color.red / 255,
             color.green / 255,
@@ -483,7 +554,7 @@ PopupSliderMenuItem.prototype = {
     _motionEvent: function(actor, event) {
         let absX, absY;
         [absX, absY] = event.get_coords();
-        this._moveHandle(absX, absY)
+        this._moveHandle(absX, absY);
         return true;
     },
 
@@ -524,8 +595,37 @@ PopupSliderMenuItem.prototype = {
         }
         return false;
     }
+};
+
+function Switch() {
+    this._init.apply(this, arguments);
 }
 
+Switch.prototype = {
+    _init: function(state) {
+        this.actor = new St.Bin({ style_class: 'toggle-switch' });
+        // Translators: this MUST be either "toggle-switch-us"
+        // (for toggle switches containing the English words
+        // "ON" and "OFF") or "toggle-switch-intl" (for toggle
+        // switches containing "◯" and "|"). Other values will
+        // simply result in invisible toggle switches.
+        this.actor.add_style_class_name(_("toggle-switch-us"));
+        this.setToggleState(state);
+    },
+
+    setToggleState: function(state) {
+        if (state)
+            this.actor.add_style_pseudo_class('checked');
+        else
+            this.actor.remove_style_pseudo_class('checked');
+        this.state = state;
+    },
+
+    toggle: function() {
+        this.setToggleState(!this.state);
+    }
+};
+
 function PopupSwitchMenuItem() {
     this._init.apply(this, arguments);
 }
@@ -533,8 +633,8 @@ function PopupSwitchMenuItem() {
 PopupSwitchMenuItem.prototype = {
     __proto__: PopupBaseMenuItem.prototype,
 
-    _init: function(text, active) {
-        PopupBaseMenuItem.prototype._init.call(this);
+    _init: function(text, active, params) {
+        PopupBaseMenuItem.prototype._init.call(this, params);
 
         this.label = new St.Label({ text: text });
         this._switch = new Switch(active);
@@ -559,18 +659,17 @@ PopupSwitchMenuItem.prototype = {
     setToggleState: function(state) {
         this._switch.setToggleState(state);
     }
-}
+};
 
-
-function PopupImageMenuItem(text, iconName) {
-    this._init(text, iconName);
+function PopupImageMenuItem() {
+    this._init.apply(this, arguments);
 }
 
 PopupImageMenuItem.prototype = {
     __proto__: PopupBaseMenuItem.prototype,
 
-    _init: function (text, iconName) {
-        PopupBaseMenuItem.prototype._init.call(this);
+    _init: function (text, iconName, params) {
+        PopupBaseMenuItem.prototype._init.call(this, params);
 
         this.label = new St.Label({ text: text });
         this.addActor(this.label);
@@ -615,8 +714,12 @@ PopupMenuBase.prototype = {
     _init: function(sourceActor, styleClass) {
         this.sourceActor = sourceActor;
 
-        this.box = new St.BoxLayout({ style_class: styleClass,
-                                      vertical: true });
+        if (styleClass !== undefined) {
+            this.box = new St.BoxLayout({ style_class: styleClass,
+                                          vertical: true });
+        } else {
+            this.box = new St.BoxLayout({ vertical: true });
+        }
 
         this.isOpen = false;
         this._activeMenuItem = null;
@@ -630,31 +733,28 @@ PopupMenuBase.prototype = {
         }));
     },
 
-    addMenuItem: function(menuItem, position) {
-        if (position == undefined)
-            this.box.add(menuItem.actor);
-        else
-            this.box.insert_actor(menuItem.actor, position);
-        if (menuItem instanceof PopupSubMenuMenuItem) {
-            if (position == undefined)
-                this.box.add(menuItem.menu.actor);
-            else
-                this.box.insert_actor(menuItem.menu.actor, position + 1);
-            menuItem._subMenuActivateId = menuItem.menu.connect('activate', Lang.bind(this, function() {
-                this.emit('activate');
-                this.close(true);
-            }));
-            menuItem._subMenuActiveChangeId = menuItem.menu.connect('active-changed', Lang.bind(this, function(submenu, submenuItem) {
-                if (this._activeMenuItem && this._activeMenuItem != submenuItem)
-                    this._activeMenuItem.setActive(false);
-                this._activeMenuItem = submenuItem;
-                this.emit('active-changed', submenuItem);
-            }));
-            menuItem._closingId = this.connect('open-state-changed', function(self, open) {
-                if (!open)
-                    menuItem.menu.close(false);
-            });
-        }
+    /**
+     * _connectSubMenuSignals:
+     * @object: a menu item, or a menu section
+     * @menu: a sub menu, or a menu section
+     *
+     * Connects to signals on @menu that are necessary for
+     * operating the submenu, and stores the ids on @object.
+     */
+    _connectSubMenuSignals: function(object, menu) {
+        object._subMenuActivateId = menu.connect('activate', Lang.bind(this, function() {
+            this.emit('activate');
+            this.close(true);
+        }));
+        object._subMenuActiveChangeId = menu.connect('active-changed', Lang.bind(this, function(submenu, submenuItem) {
+            if (this._activeMenuItem && this._activeMenuItem != submenuItem)
+                this._activeMenuItem.setActive(false);
+            this._activeMenuItem = submenuItem;
+            this.emit('active-changed', submenuItem);
+        }));
+    },
+
+    _connectItemSignals: function(menuItem) {
         menuItem._activeChangeId = menuItem.connect('active-changed', Lang.bind(this, function (menuItem, active) {
             if (active && this._activeMenuItem != menuItem) {
                 if (this._activeMenuItem)
@@ -683,6 +783,41 @@ PopupMenuBase.prototype = {
         }));
     },
 
+    addMenuItem: function(menuItem, position) {
+        let before_item = null;
+        if (position == undefined) {
+            this.box.add(menuItem.actor);
+        } else {
+            let items = this._getMenuItems();
+            if (position < items.length) {
+                before_item = items[position].actor;
+                this.box.insert_before(menuItem.actor, before_item);
+            } else
+                this.box.add(menuItem.actor);
+        }
+        if (menuItem instanceof PopupMenuSection) {
+            this._connectSubMenuSignals(menuItem, menuItem);
+            menuItem.connect('destroy', Lang.bind(this, function() {
+                menuItem.disconnect(menuItem._subMenuActivateId);
+                menuItem.disconnect(menuItem._subMenuActiveChangeId);
+            }));
+        } else if (menuItem instanceof PopupSubMenuMenuItem) {
+            if (before_item == null)
+                this.box.add(menuItem.menu.actor);
+            else
+                this.box.insert_before(menuItem.menu.actor, before_item);
+            this._connectSubMenuSignals(menuItem, menuItem.menu);
+            this._connectItemSignals(menuItem);
+            menuItem._closingId = this.connect('open-state-changed', function(self, open) {
+                if (!open)
+                    menuItem.menu.close(false);
+            });
+        } else if (menuItem instanceof PopupBaseMenuItem)
+            this._connectItemSignals(menuItem);
+        else
+            throw TypeError("Invalid argument to PopupMenuBase.addMenuItem()");
+    },
+
     getColumnWidths: function() {
         let columnWidths = [];
         let items = this.box.get_children();
@@ -710,12 +845,16 @@ PopupMenuBase.prototype = {
         this.box.add(actor);
     },
 
-    getMenuItems: function() {
-        return this.box.get_children().map(function (actor) { return actor._delegate; }).filter(function(item) { return item instanceof PopupBaseMenuItem; });
+    _getMenuItems: function() {
+        return this.box.get_children().map(function (actor) {
+            return actor._delegate;
+        }).filter(function(item) {
+            return item instanceof PopupBaseMenuItem || item instanceof PopupMenuSection;
+        });
     },
 
     removeAll: function() {
-        let children = this.getMenuItems();
+        let children = this._getMenuItems();
         for (let i = 0; i < children.length; i++) {
             let item = children[i];
             item.destroy();
@@ -926,6 +1065,34 @@ PopupSubMenu.prototype = {
     }
 };
 
+/**
+ * PopupMenuSection:
+ *
+ * A section of a PopupMenu which is handled like a submenu
+ * (you can add and remove items, you can destroy it, you
+ * can add it to another menu), but is completely transparent
+ * to the user
+ */
+function PopupMenuSection() {
+    this._init.apply(this, arguments);
+}
+
+PopupMenuSection.prototype = {
+    __proto__: PopupMenuBase.prototype,
+
+    _init: function() {
+        PopupMenuBase.prototype._init.call(this);
+
+        this.actor = this.box;
+        this.actor._delegate = this;
+        this.isOpen = true;
+    },
+
+    // deliberately ignore any attempt to open() or close()
+    open: function(animate) { },
+    close: function() { },
+}
+
 function PopupSubMenuMenuItem() {
     this._init.apply(this, arguments);
 }
@@ -1187,6 +1354,18 @@ PopupMenuManager.prototype = {
         return -1;
     },
 
+    _nextMenu: function(pos, direction) {
+        for (let i = 1; i < this._menus.length; i++) {
+            let candidate = mod(pos + i * direction, this._menus.length);
+            let menu = this._menus[candidate].menu;
+            if (!menu.sourceActor || menu.sourceActor.visible)
+                return menu;
+        }
+        // no menu is found? this should not happen
+        // anyway stay on current menu
+        return this._menus[pos];
+    },
+
     _onEventCapture: function(actor, event) {
         if (!this.grabbed)
             return false;
@@ -1225,7 +1404,7 @@ PopupMenuManager.prototype = {
         if (symbol == Clutter.Left || symbol == Clutter.Right) {
             let direction = symbol == Clutter.Right ? 1 : -1;
             let pos = this._findMenu(this._activeMenu);
-            let next = this._menus[mod(pos + direction, this._menus.length)].menu;
+            let next = this._nextMenu(pos, direction);
             if (next != this._activeMenu) {
                 this._changeMenu(next);
                 next.activateFirst();
