@@ -267,7 +267,7 @@ OpenSearchSystem.prototype = {
         return lang != null;
     },
 
-    activateResult: function(id) {
+    activateResult: function(id, params) {
         let searchTerms = this._terms.join(' ');
 
         let url = this._providers[id].url.replace('{searchTerms}', encodeURIComponent(searchTerms));
@@ -355,7 +355,7 @@ SearchSystem.prototype = {
     updateSearch: function(searchString) {
         searchString = searchString.replace(/^\s+/g, '').replace(/\s+$/g, '');
         if (searchString == '')
-            return null;
+            return [];
 
         let terms = searchString.split(/\s+/);
         let isSubSearch = terms.length == this._previousTerms.length;
