@@ -5,6 +5,7 @@ const GLib = imports.gi.GLib;
 const Lang = imports.lang;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
+const UPowerGlib = imports.gi.UPowerGlib;
 const Gettext = imports.gettext.domain('gnome-shell');
 const _ = Gettext.gettext;
 
@@ -12,6 +13,7 @@ const GnomeSession = imports.misc.gnomeSession;
 const Main = imports.ui.main;
 const PanelMenu = imports.ui.panelMenu;
 const PopupMenu = imports.ui.popupMenu;
+const Util = imports.misc.util;
 
 // Adapted from gdm/gui/user-switch-applet/applet.c
 //
@@ -26,17 +28,19 @@ StatusMenuButton.prototype = {
     __proto__: PanelMenu.Button.prototype,
 
     _init: function() {
-        PanelMenu.Button.prototype._init.call(this, St.Align.START);
+        PanelMenu.Button.prototype._init.call(this, 0.0);
         let box = new St.BoxLayout({ name: 'panelStatusMenu' });
         this.actor.set_child(box);
 
         this._gdm = Gdm.UserManager.ref_default();
-        this._gdm.queue_load()
+        this._gdm.queue_load();
 
         this._user = this._gdm.get_user(GLib.get_user_name());
         this._presence = new GnomeSession.Presence();
         this._presenceItems = {};
 
+        this._upClient = new UPowerGlib.Client();
+
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
         this._iconBox = new St.Bin();
@@ -60,6 +64,8 @@ StatusMenuButton.prototype = {
         this._gdm.connect('notify::is-loaded', Lang.bind(this, this._updateSwitchUser));
         this._gdm.connect('user-added', Lang.bind(this, this._updateSwitchUser));
         this._gdm.connect('user-removed', Lang.bind(this, this._updateSwitchUser));
+
+        this._upClient.connect('notify::can-suspend', Lang.bind(this, this._updateSuspendOrPowerOff));
     },
 
     _onDestroy: function() {
@@ -81,6 +87,21 @@ StatusMenuButton.prototype = {
             this._loginScreenItem.actor.hide();
     },
 
+    _updateSuspendOrPowerOff: function() {
+        this._haveSuspend = this._upClient.get_can_suspend();
+
+        if (!this._suspendOrPowerOffItem)
+            return;
+
+        // If we can't suspend show Power Off... instead
+        // and disable the alt key
+        if (!this._haveSuspend) {
+            this._suspendOrPowerOffItem.updateText(_("Power Off..."), null);
+        } else {
+            this._suspendOrPowerOffItem.updateText(_("Suspend"), ("Power Off..."));
+        }
+    },
+
     _updatePresenceIcon: function(presence, status) {
         if (status == GnomeSession.PresenceStatus.AVAILABLE)
             this._iconBox.child = this._availableIcon;
@@ -98,12 +119,12 @@ StatusMenuButton.prototype = {
     _createSubMenu: function() {
         let item;
 
-        item = new PopupMenu.PopupImageMenuItem(_("Available"), 'user-available', true);
+        item = new PopupMenu.PopupImageMenuItem(_("Available"), 'user-available');
         item.connect('activate', Lang.bind(this, this._setPresenceStatus, GnomeSession.PresenceStatus.AVAILABLE));
         this.menu.addMenuItem(item);
         this._presenceItems[GnomeSession.PresenceStatus.AVAILABLE] = item;
 
-        item = new PopupMenu.PopupImageMenuItem(_("Busy"), 'user-busy', true);
+        item = new PopupMenu.PopupImageMenuItem(_("Busy"), 'user-busy');
         item.connect('activate', Lang.bind(this, this._setPresenceStatus, GnomeSession.PresenceStatus.BUSY));
         this.menu.addMenuItem(item);
         this._presenceItems[GnomeSession.PresenceStatus.BUSY] = item;
@@ -138,13 +159,12 @@ StatusMenuButton.prototype = {
         item = new PopupMenu.PopupSeparatorMenuItem();
         this.menu.addMenuItem(item);
 
-        item = new PopupMenu.PopupMenuItem(_("Suspend..."));
-        item.connect('activate', Lang.bind(this, this._onShutDownActivate));
-        this.menu.addMenuItem(item);
-
-        item = new PopupMenu.PopupMenuItem(_("Shut Down..."));
-        item.connect('activate', Lang.bind(this, this._onShutDownActivate));
+        item = new PopupMenu.PopupAlternatingMenuItem(_("Suspend"),
+                                                      _("Power Off..."));
         this.menu.addMenuItem(item);
+        this._suspendOrPowerOffItem = item;
+        item.connect('activate', Lang.bind(this, this._onSuspendOrPowerOffActivate));
+        this._updateSuspendOrPowerOff();
     },
 
     _setPresenceStatus: function(item, event, status) {
@@ -153,17 +173,17 @@ StatusMenuButton.prototype = {
 
     _onMyAccountActivate: function() {
         Main.overview.hide();
-        this._spawn(['gnome-control-center', 'user-accounts']);
+        Util.spawnDesktop('gnome-user-accounts-panel');
     },
 
     _onPreferencesActivate: function() {
         Main.overview.hide();
-        this._spawn(['gnome-control-center']);
+        Util.spawnDesktop('gnome-control-center');
     },
 
     _onLockScreenActivate: function() {
         Main.overview.hide();
-        this._spawn(['gnome-screensaver-command', '--lock']);
+        Util.spawn(['gnome-screensaver-command', '--lock']);
     },
 
     _onLoginScreenActivate: function() {
@@ -174,19 +194,17 @@ StatusMenuButton.prototype = {
 
     _onQuitSessionActivate: function() {
         Main.overview.hide();
-        this._spawn(['gnome-session-save', '--logout-dialog']);
+        Util.spawn(['gnome-session-save', '--logout-dialog']);
     },
 
-    _onShutDownActivate: function() {
+    _onSuspendOrPowerOffActivate: function() {
         Main.overview.hide();
-        this._spawn(['gnome-session-save', '--shutdown-dialog']);
-    },
 
-    _spawn: function(args) {
-        // FIXME: once Shell.Process gets support for signalling
-        // errors we should pop up an error dialog or something here
-        // on failure
-        let p = new Shell.Process({'args' : args});
-        p.run();
+        if (this._haveSuspend &&
+            this._suspendOrPowerOffItem.state == PopupMenu.PopupAlternatingMenuItemState.DEFAULT) {
+            this._upClient.suspend_sync(null);
+        } else {
+            Util.spawn(['gnome-session-save', '--shutdown-dialog']);
+        }
     }
 };
