@@ -23,22 +23,28 @@ function SearchEntry(focusBase) {
 SearchEntry.prototype = {
     _init : function(focusBase) {
         this.actor = new St.Entry({ name: 'searchEntry',
-                                    hint_text: _("Search your computer") });
+                                    /* Translators: this is the text displayed
+                                       in the search entry when no search is
+                                       active; it should not exceed ~30
+                                       characters */
+                                    hint_text: _("Type to search..."),
+                                    track_hover: true });
         this.entry = this.actor.clutter_text;
 
-        this.actor.clutter_text.connect('text-changed', Lang.bind(this,
-            function() {
-                if (this.isActive())
-                    this.actor.set_secondary_icon_from_file(global.imagedir +
-                                                            'close-black.svg');
-                else
-                    this.actor.set_secondary_icon_from_file(null);
-            }));
-        this.actor.connect('secondary-icon-clicked', Lang.bind(this,
-            function() {
-                this.reset();
-            }));
+        this._inactiveIcon = new St.Icon({ style_class: 'search-entry-icon',
+                                           icon_name: 'edit-find',
+                                           icon_type: St.IconType.SYMBOLIC });
+        this._activeIcon = new St.Icon({ style_class: 'search-entry-icon',
+                                         icon_name: 'edit-clear',
+                                         icon_type: St.IconType.SYMBOLIC });
+        this.actor.set_secondary_icon(this._inactiveIcon);
+
+        this._iconClickedId = 0;
+
+        this.actor.clutter_text.connect('text-changed',
+                                        Lang.bind(this, this._onTextChanged));
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
+        this.actor.connect('notify::mapped', Lang.bind(this, this._onMapped));
 
         global.stage.connect('notify::key-focus', Lang.bind(this, this._updateCursorVisibility));
 
@@ -56,29 +62,23 @@ SearchEntry.prototype = {
             this.entry.set_cursor_visible(false);
     },
 
-    show: function() {
-        if (this._capturedEventId == 0)
+    _onMapped: function() {
+        if (this.actor.mapped) {
+            // Enable 'find-as-you-type'
             this._capturedEventId = global.stage.connect('captured-event',
                                  Lang.bind(this, this._onCapturedEvent));
-        this.entry.set_cursor_visible(true);
-        this.entry.set_selection(0, 0);
-    },
-
-    hide: function() {
-        if (this._capturedEventId > 0) {
-            global.stage.disconnect(this._capturedEventId);
+            this.entry.set_cursor_visible(true);
+            this.entry.set_selection(0, 0);
+        } else {
+            // Disable 'find-as-you-type'
+            if (this._capturedEventId > 0)
+                global.stage.disconnect(this._capturedEventId);
             this._capturedEventId = 0;
         }
     },
 
     reset: function () {
-        let [x, y, mask] = global.get_pointer();
-        let actor = global.stage.get_actor_at_pos (Clutter.PickMode.REACTIVE,
-                                                   x, y);
-        // this.actor is never hovered directly, only its clutter_text and icon
-        let hovered = this.actor == actor.get_parent();
-
-        this.actor.set_hover(hovered);
+        this.actor.sync_hover();
 
         this.entry.text = '';
 
@@ -157,11 +157,32 @@ SearchEntry.prototype = {
         }
     },
 
+    _onTextChanged: function() {
+        if (this.isActive()) {
+            this.actor.set_secondary_icon(this._activeIcon);
+
+            if (this._iconClickedId == 0)
+            this._iconClickedId = this.actor.connect('secondary-icon-clicked',
+                Lang.bind(this, function() {
+                    this.reset();
+                }));
+        } else {
+            if (this._iconClickedId > 0)
+                this.actor.disconnect(this._iconClickedId);
+            this._iconClickedId = 0;
+
+            this.actor.set_secondary_icon(this._inactiveIcon);
+        }
+    },
+
     _onDestroy: function() {
         if (this._capturedEventId > 0) {
             global.stage.disconnect(this._capturedEventId);
             this._capturedEventId = 0;
         }
+
+        this._activeIcon = null;
+        this._inactiveIcon = null;
     }
 };
 Signals.addSignalMethods(SearchEntry.prototype);
@@ -266,13 +287,6 @@ SearchTab.prototype = {
         }));
     },
 
-    setFindAsYouType: function(enabled) {
-        if (enabled)
-            this._searchEntry.show();
-        else
-            this._searchEntry.hide();
-    },
-
     show: function() {
         BaseTab.prototype.show.call(this);
 
@@ -356,6 +370,8 @@ ViewSelector.prototype = {
     _init : function() {
         this.actor = new St.BoxLayout({ name: 'viewSelector',
                                         vertical: true });
+        this.actor.connect('key-press-event',
+                           Lang.bind(this, this._onKeyPress));
 
         // The tab bar is located at the top of the view selector and
         // holds both "normal" tab labels and the search entry. The former
@@ -398,9 +414,12 @@ ViewSelector.prototype = {
                 this._switchTab(this._activeTab);
             }));
 
-        this._keyPressId = 0;
-        this._itemDragBeginId = 0;
-        this._overviewHidingId = 0;
+        Main.overview.connect('item-drag-begin',
+                              Lang.bind(this, this._switchDefaultTab));
+        Main.overview.connect('showing',
+                              Lang.bind(this, this._switchDefaultTab));
+        Main.overview.connect('hiding',
+                              Lang.bind(this, this._switchDefaultTab));
 
         // Public constraints which may be used to tie actors' height or
         // vertical position to the current tab's content; as the content's
@@ -577,41 +596,6 @@ ViewSelector.prototype = {
 
     addSearchProvider: function(provider) {
         this._searchTab.addSearchProvider(provider);
-    },
-
-    show: function() {
-        this._searchTab.setFindAsYouType(true);
-
-        if (this._itemDragBeginId == 0)
-            this._itemDragBeginId = Main.overview.connect('item-drag-begin',
-                                                          Lang.bind(this, this._switchDefaultTab));
-        if (this._overviewHidingId == 0)
-            this._overviewHidingId = Main.overview.connect('hiding',
-                                                           Lang.bind(this, this._switchDefaultTab));
-        if (this._keyPressId == 0)
-            this._keyPressId = this.actor.connect('key-press-event',
-                                                  Lang.bind(this, this._onKeyPress));
-
-        this._switchDefaultTab();
-    },
-
-    hide: function() {
-        this._searchTab.setFindAsYouType(false);
-
-        if (this._keyPressId > 0) {
-            this.actor.disconnect(this._keyPressId);
-            this._keyPressId = 0;
-        }
-
-        if (this._itemDragBeginId > 0) {
-            Main.overview.disconnect(this._itemDragBeginId);
-            this._itemDragBeginId = 0;
-        }
-
-        if (this._overviewHidingId > 0) {
-            Main.overview.disconnect(this._overviewHidingId);
-            this._overviewHidingId = 0;
-        }
     }
 };
 Signals.addSignalMethods(ViewSelector.prototype);
