@@ -20,9 +20,6 @@ const FOCUS_ANIMATION_TIME = 0.15;
 
 const WINDOW_DND_SIZE = 256;
 
-const FRAME_COLOR = new Clutter.Color();
-FRAME_COLOR.from_pixel(0xffffffff);
-
 const SCROLL_SCALE_AMOUNT = 100 / 5;
 
 const LIGHTBOX_FADE_TIME = 0.1;
@@ -54,11 +51,6 @@ function _clamp(value, min, max) {
     return Math.max(min, Math.min(max, value));
 }
 
-// Spacing between workspaces. At the moment, the same spacing is used
-// in both zoomed-in and zoomed-out views; this is slightly
-// metaphor-breaking, but the alternatives are also weird.
-const GRID_SPACING = 15;
-const FRAME_SIZE = GRID_SPACING / 3;
 
 function ScaledPoint(x, y, scaleX, scaleY) {
     [this.x, this.y, this.scaleX, this.scaleY] = arguments;
@@ -112,6 +104,12 @@ WindowClone.prototype = {
 
         this._stackAbove = null;
 
+        this._sizeChangedId = this.realWindow.connect('size-changed', Lang.bind(this, function() {
+            this.emit('size-changed');
+        }));
+        this._realWindowDestroyId = this.realWindow.connect('destroy',
+            Lang.bind(this, this._disconnectRealWindowSignals));
+
         this.actor.connect('button-release-event',
                            Lang.bind(this, this._onButtonRelease));
 
@@ -140,7 +138,10 @@ WindowClone.prototype = {
         if (this.inDrag || this._zooming)
             // We'll fix up the stack after the drag/zooming
             return;
-        this.actor.raise(this._stackAbove);
+        if (this._stackAbove == null)
+            this.actor.lower_bottom();
+        else
+            this.actor.raise(this._stackAbove);
     },
 
     destroy: function () {
@@ -161,7 +162,19 @@ WindowClone.prototype = {
         }
     },
 
+    _disconnectRealWindowSignals: function() {
+        if (this._sizeChangedId > 0)
+            this.realWindow.disconnect(this._sizeChangedId);
+        this._sizeChangedId = 0;
+
+        if (this._realWindowDestroyId > 0)
+            this.realWindow.disconnect(this._realWindowDestroyId);
+        this._realWindowDestroyId = 0;
+    },
+
     _onDestroy: function() {
+        this._disconnectRealWindowSignals();
+
         this.metaWindow._delegate = null;
         this.actor._delegate = null;
         if (this._zoomLightbox)
@@ -247,11 +260,13 @@ WindowClone.prototype = {
         this.emit('zoom-end');
 
         this.actor.reparent(this._origParent);
+        if (this._stackAbove == null)
+            this.actor.lower_bottom();
         // If the workspace has been destroyed while we were reparented to
         // the stage, _stackAbove will be unparented and we can't raise our
         // actor above it - as we are bound to be destroyed anyway in that
         // case, we can skip that step
-        if (this._stackAbove && this._stackAbove.get_parent())
+        else if (this._stackAbove.get_parent())
             this.actor.raise(this._stackAbove);
 
         [this.actor.x, this.actor.y]             = this._zoomLocalOrig.getPosition();
@@ -283,82 +298,20 @@ WindowClone.prototype = {
         // We may not have a parent if DnD completed successfully, in
         // which case our clone will shortly be destroyed and replaced
         // with a new one on the target workspace.
-        if (this.actor.get_parent() != null)
-            this.actor.raise(this._stackAbove);
+        if (this.actor.get_parent() != null) {
+            if (this._stackAbove == null)
+                this.actor.lower_bottom();
+            else
+                this.actor.raise(this._stackAbove);
+        }
+
 
         this.emit('drag-end');
     }
 };
-
 Signals.addSignalMethods(WindowClone.prototype);
 
 
-function DesktopClone(window) {
-    this._init(window);
-}
-
-DesktopClone.prototype = {
-    _init : function(window) {
-        this.actor = new Clutter.Group({ reactive: true });
-
-        let background = new Clutter.Clone({ source: Main.background.source });
-        this.actor.add_actor(background);
-
-        if (window) {
-            this._desktop = new Clutter.Clone({ source: window.get_texture() });
-            this.actor.add_actor(this._desktop);
-            this._desktop.hide();
-        } else {
-            this._desktop = null;
-        }
-
-        this.actor.connect('button-release-event',
-                           Lang.bind(this, this._onButtonRelease));
-    },
-
-    zoomFromOverview: function(fadeInIcons) {
-        if (this._desktop == null)
-            return;
-
-        if (fadeInIcons) {
-            this._desktop.opacity = 0;
-            this._desktop.show();
-            Tweener.addTween(this._desktop,
-                             { opacity: 255,
-                               time: Overview.ANIMATION_TIME,
-                               transition: 'easeOutQuad' });
-        }
-    },
-
-    zoomToOverview: function(fadeOutIcons) {
-        if (this._desktop == null)
-            return;
-
-        if (fadeOutIcons) {
-            this._desktop.opacity = 255;
-            this._desktop.show();
-            Tweener.addTween(this._desktop,
-                             { opacity: 0,
-                               time: Overview.ANIMATION_TIME,
-                               transition: 'easeOutQuad',
-                               onComplete: Lang.bind(this,
-                                   function() {
-                                       this._desktop.hide();
-                                   })
-                             });
-        } else {
-            this._desktop.hide();
-        }
-    },
-
-    _onButtonRelease : function (actor, event) {
-        this.emit('selected', event.get_time());
-    }
-};
-
-Signals.addSignalMethods(DesktopClone.prototype);
-
-
 /**
  * @windowClone: Corresponding window clone
  * @parentActor: The actor which will be the parent of all overlay items
@@ -461,7 +414,7 @@ WindowOverlay.prototype = {
     // These parameters are not the values retrieved with
     // get_transformed_position() and get_transformed_size(),
     // as windowClone might be moving.
-    // See Workspace._fadeInWindowOverlay
+    // See Workspace._showWindowOverlay
     updatePositions: function(cloneX, cloneY, cloneWidth, cloneHeight) {
         let button = this.closeButton;
         let title = this.title;
@@ -534,12 +487,10 @@ WindowOverlay.prototype = {
 
     _idleToggleCloseButton: function() {
         this._idleToggleCloseId = 0;
-        let [x, y, mask] = global.get_pointer();
-        let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE,
-                                                  x, y);
-        if (actor != this._windowClone.actor && actor != this.closeButton) {
+        if (!this._windowClone.actor.has_pointer &&
+            !this.closeButton.has_pointer)
             this.closeButton.hide();
-        }
+
         return false;
     },
 
@@ -561,7 +512,6 @@ WindowOverlay.prototype = {
         this._parentActor.queue_relayout();
     }
 };
-
 Signals.addSignalMethods(WindowOverlay.prototype);
 
 const WindowPositionFlags = {
@@ -602,46 +552,18 @@ Workspace.prototype = {
         this.actor.height = global.screen_height;
         this.scale = 1.0;
 
-        let windows = global.get_window_actors().filter(this._isMyWindow, this);
+        let windows = Main.getWindowActorsForWorkspace(this.metaWorkspace.index());
 
-        // Find the desktop window
-        for (let i = 0; i < windows.length; i++) {
-            if (windows[i].meta_window.get_window_type() == Meta.WindowType.DESKTOP) {
-                this._desktop = new DesktopClone(windows[i]);
-                break;
-            }
-        }
-        // If there wasn't one, fake it
-        if (!this._desktop)
-            this._desktop = new DesktopClone();
-
-        this._desktop.connect('selected',
-                              Lang.bind(this,
-                                        function(clone, time) {
-                                            // Only switch to the workspace when there's no application windows
-                                            // open (we always have one window for the desktop).  The problem
-                                            // is that it's too easy to miss an app window and get the wrong
-                                            // one focused.
-                                            if (this._windows.length == 1) {
-                                                this.metaWorkspace.activate(time);
-                                                Main.overview.hide();
-                                            }
-                                        }));
-        this.actor.add_actor(this._desktop.actor);
-
-        // Create clones for remaining windows that should be
+        // Create clones for windows that should be
         // visible in the Overview
-        this._windows = [this._desktop];
-        this._windowOverlays = [ null ];
+        this._windows = [];
+        this._windowOverlays = [];
         for (let i = 0; i < windows.length; i++) {
             if (this._isOverviewWindow(windows[i])) {
                 this._addWindowClone(windows[i]);
             }
         }
 
-        // A filter for what windows we display
-        this._showOnlyWindows = null;
-
         // Track window changes
         this._windowAddedId = this.metaWorkspace.connect('window-added',
                                                           Lang.bind(this, this._windowAdded));
@@ -649,10 +571,6 @@ Workspace.prototype = {
                                                             Lang.bind(this, this._windowRemoved));
         this._repositionWindowsId = 0;
 
-        this._visible = false;
-
-        this._frame = null;
-
         this.leavingOverview = false;
     },
 
@@ -676,91 +594,12 @@ Workspace.prototype = {
             new_parent.add_actor(this._windowOverlaysGroup);
     },
 
-    /**
-     * lookupCloneForMetaWindow:
-     * @metaWindow: A #MetaWindow
-     *
-     * Given a #MetaWindow instance, find the WindowClone object
-     * which represents it in the workspaces display.
-     */
-    lookupCloneForMetaWindow: function (metaWindow) {
-        let index = this._lookupIndex (metaWindow);
-        return index < 0 ? null : this._windows[index];
-    },
-
     containsMetaWindow: function (metaWindow) {
         return this._lookupIndex(metaWindow) >= 0;
     },
 
-    setShowOnlyWindows: function(showOnlyWindows, reposition) {
-        this._showOnlyWindows = showOnlyWindows;
-        this._resetCloneVisibility();
-        if (reposition)
-            this.positionWindows(WindowPositionFlags.ANIMATE);
-    },
-
-    /**
-     * setLightboxMode:
-     * @showLightbox: If true, dim background and allow highlighting a specific window
-     *
-     * This function also resets the highlighted window state.
-     */
-    setLightboxMode: function (showLightbox) {
-        if (!this._lightbox)
-            this._lightbox = new Lightbox.Lightbox(this.actor,
-                                                   { fadeTime: LIGHTBOX_FADE_TIME });
-
-        if (showLightbox)
-            this._lightbox.show();
-        else
-            this._lightbox.hide();
-
-        if (this._frame)
-            this._frame.set_opacity(showLightbox ? 150 : 255);
-    },
-
-    /**
-     * setHighlightWindow:
-     * @metaWindow: A #MetaWindow
-     *
-     * Draw the user's attention to the given window @metaWindow.
-     */
-    setHighlightWindow: function (metaWindow) {
-        if (!this._lightbox)
-            return;
-
-        let actor;
-        if (metaWindow != null) {
-            let clone = this.lookupCloneForMetaWindow(metaWindow);
-            actor = clone.actor;
-        }
-        this._lightbox.highlight(actor);
-    },
-
-    // Mark the workspace selected/not-selected
-    setSelected : function(selected) {
-        // Don't draw a frame if we only have one workspace
-        if (selected && global.screen.n_workspaces > 1) {
-            if (this._frame)
-                return;
-
-            // FIXME: do something cooler-looking using clutter-cairo
-            this._frame = new Clutter.Rectangle({ color: FRAME_COLOR });
-            this.actor.add_actor(this._frame);
-            this._frame.set_position(this._desktop.actor.x - FRAME_SIZE / this.actor.scale_x,
-                                     this._desktop.actor.y - FRAME_SIZE / this.actor.scale_y);
-            this._frame.set_size(this._desktop.actor.width + 2 * FRAME_SIZE / this.actor.scale_x,
-                                 this._desktop.actor.height + 2 * FRAME_SIZE / this.actor.scale_y);
-            this._frame.lower_bottom();
-
-            this._framePosHandler = this.actor.connect('notify::scale-x', Lang.bind(this, this._updateFramePosition));
-        } else {
-            if (!this._frame)
-                return;
-            this.actor.disconnect(this._framePosHandler);
-            this._frame.destroy();
-            this._frame = null;
-        }
+    isEmpty: function() {
+        return this._windows.length == 0;
     },
 
     /**
@@ -770,55 +609,7 @@ Workspace.prototype = {
      * Set the workspace (desktop) reactive
      **/
     setReactive: function(reactive) {
-        this._desktop.actor.reactive = reactive;
-    },
-
-    _updateFramePosition : function() {
-        this._frame.set_position(this._desktop.actor.x - FRAME_SIZE / this.actor.scale_x,
-                                 this._desktop.actor.y - FRAME_SIZE / this.actor.scale_y);
-        this._frame.set_size(this._desktop.actor.width + 2 * FRAME_SIZE / this.actor.scale_x,
-                             this._desktop.actor.height + 2 * FRAME_SIZE / this.actor.scale_y);
-    },
-
-    _isCloneVisible: function(clone) {
-        return this._showOnlyWindows == null || (clone.metaWindow in this._showOnlyWindows);
-    },
-
-    /**
-     * _getVisibleClones:
-     *
-     * Returns a list WindowClone objects where the clone isn't filtered
-     * out by any application filter.  The clone for the desktop is excluded.
-     * The returned array will always be newly allocated; it is not in any
-     * defined order, and thus it's convenient to call .sort() with your
-     * choice of sorting function.
-     */
-    _getVisibleClones: function() {
-        let visible = [];
-
-        for (let i = 1; i < this._windows.length; i++) {
-            let clone = this._windows[i];
-
-            if (!this._isCloneVisible(clone))
-                continue;
-
-            visible.push(clone);
-        }
-        return visible;
-    },
-
-    _resetCloneVisibility: function () {
-        for (let i = 1; i < this._windows.length; i++) {
-            let clone = this._windows[i];
-            let overlay = this._windowOverlays[i];
-
-            if (!this._isCloneVisible(clone)) {
-                clone.actor.hide();
-                overlay.hide();
-            } else {
-                clone.actor.show();
-            }
-        }
+        this.actor.reactive = reactive;
     },
 
     // Only use this for n <= 20 say
@@ -1002,14 +793,13 @@ Workspace.prototype = {
 
         xCenter = xCenter * global.screen_width;
 
-        let rect = new Meta.Rectangle();
-        metaWindow.get_outer_rect(rect);
+        let rect = metaWindow.get_outer_rect();
         let buttonOuterHeight, captionHeight;
         let buttonOuterWidth = 0;
 
-        if (this._windowOverlays[1]) {
-            [buttonOuterHeight, captionHeight] = this._windowOverlays[1].chromeHeights();
-            buttonOuterWidth = this._windowOverlays[1].chromeWidth() / this.scale;
+        if (this._windowOverlays[0]) {
+            [buttonOuterHeight, captionHeight] = this._windowOverlays[0].chromeHeights();
+            buttonOuterWidth = this._windowOverlays[0].chromeWidth() / this.scale;
         } else
             [buttonOuterHeight, captionHeight] = [0, 0];
         buttonOuterHeight /= this.scale;
@@ -1033,6 +823,9 @@ Workspace.prototype = {
     },
 
     setReservedSlot: function(clone) {
+        if (this._reservedSlot == clone)
+            return;
+
         if (clone && this.containsMetaWindow(clone.metaWindow)) {
             this._reservedSlot = null;
             this.positionWindows(WindowPositionFlags.ANIMATE);
@@ -1057,22 +850,23 @@ Workspace.prototype = {
             this._repositionWindowsId = 0;
         }
 
-        let totalVisible = 0;
-
-        let visibleClones = this._getVisibleClones();
+        let clones = this._windows.slice();
         if (this._reservedSlot)
-            visibleClones.push(this._reservedSlot);
+            clones.push(this._reservedSlot);
 
         let workspaceZooming = flags & WindowPositionFlags.ZOOM;
         let animate = flags & WindowPositionFlags.ANIMATE;
 
         // Start the animations
-        let slots = this._computeAllWindowSlots(visibleClones.length);
-        visibleClones = this._orderWindowsByMotionAndStartup(visibleClones, slots);
+        let slots = this._computeAllWindowSlots(clones.length);
+        clones = this._orderWindowsByMotionAndStartup(clones, slots);
 
-        for (let i = 0; i < visibleClones.length; i++) {
+        let currentWorkspace = global.screen.get_active_workspace();
+        let isOnCurrentWorkspace = this.metaWorkspace == currentWorkspace;
+
+        for (let i = 0; i < clones.length; i++) {
             let slot = slots[i];
-            let clone = visibleClones[i];
+            let clone = clones[i];
             let metaWindow = clone.metaWindow;
             let mainIndex = this._lookupIndex(metaWindow);
             let overlay = this._windowOverlays[mainIndex];
@@ -1086,7 +880,7 @@ Workspace.prototype = {
 
             if (overlay)
                 overlay.hide();
-            if (animate) {
+            if (animate && isOnCurrentWorkspace) {
                 if (!metaWindow.showing_on_its_workspace()) {
                     /* Hidden windows should fade in and grow
                      * therefore we need to resize them now so they
@@ -1116,36 +910,34 @@ Workspace.prototype = {
                                    time: Overview.ANIMATION_TIME,
                                    transition: 'easeOutQuad',
                                    onComplete: Lang.bind(this, function() {
-                                      this._fadeInWindowOverlay(clone, overlay);
+                                      this._showWindowOverlay(clone, overlay, true);
                                    })
                                  });
             } else {
                 clone.actor.set_position(x, y);
                 clone.actor.set_scale(scale, scale);
-                this._fadeInWindowOverlay(clone, overlay);
+                this._showWindowOverlay(clone, overlay, isOnCurrentWorkspace);
             }
         }
     },
 
     syncStacking: function(stackIndices) {
-        let desktopClone = this._windows[0];
+        let clones = this._windows.slice();
+        clones.sort(function (a, b) { return stackIndices[a.metaWindow.get_stable_sequence()] - stackIndices[b.metaWindow.get_stable_sequence()]; });
 
-        let visibleClones = this._getVisibleClones();
-        visibleClones.sort(function (a, b) { return stackIndices[a.metaWindow.get_stable_sequence()] - stackIndices[b.metaWindow.get_stable_sequence()]; });
-
-        for (let i = 0; i < visibleClones.length; i++) {
-            let clone = visibleClones[i];
+        for (let i = 0; i < clones.length; i++) {
+            let clone = clones[i];
             let metaWindow = clone.metaWindow;
             if (i == 0) {
-                clone.setStackAbove(desktopClone.actor);
+                clone.setStackAbove(null);
             } else {
-                let previousClone = visibleClones[i - 1];
+                let previousClone = clones[i - 1];
                 clone.setStackAbove(previousClone.actor);
             }
         }
     },
 
-    _fadeInWindowOverlay: function(clone, overlay) {
+    _showWindowOverlay: function(clone, overlay, fade) {
         if (clone.inDrag)
             return;
 
@@ -1157,34 +949,26 @@ Workspace.prototype = {
         // be after the workspace animation finishes.
         let [cloneX, cloneY] = clone.actor.get_position();
         let [cloneWidth, cloneHeight] = clone.actor.get_size();
-        cloneX = this.gridX + this.scale * cloneX;
-        cloneY = this.gridY + this.scale * cloneY;
+        cloneX = this.x + this.scale * cloneX;
+        cloneY = this.y + this.scale * cloneY;
         cloneWidth = this.scale * clone.actor.scale_x * cloneWidth;
         cloneHeight = this.scale * clone.actor.scale_y * cloneHeight;
 
-        if (!this._windowOverlaysGroup.visible)
-            this._windowOverlaysGroup.show();
-
         if (overlay) {
             overlay.updatePositions(cloneX, cloneY, cloneWidth, cloneHeight);
-            overlay.fadeIn();
+            if (fade)
+                overlay.fadeIn();
+            else
+                overlay.show();
         }
     },
 
-    _fadeInAllOverlays: function() {
-        for (let i = 1; i < this._windows.length; i++) {
+    _showAllOverlays: function() {
+        let currentWorkspace = global.screen.get_active_workspace();
+        for (let i = 0; i < this._windows.length; i++) {
             let clone = this._windows[i];
             let overlay = this._windowOverlays[i];
-            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
-                continue;
-            this._fadeInWindowOverlay(clone, overlay);
-        }
-    },
-
-    _hideAllOverlays: function() {
-        for (let i = 1; i< this._windows.length; i++) {
-            let overlay = this._windowOverlays[i];
-            overlay.hide();
+            this._showWindowOverlay(clone, overlay, this.metaWorkspace == currentWorkspace);
         }
     },
 
@@ -1197,8 +981,8 @@ Workspace.prototype = {
         let wsHeight = this.actor.height * this.scale;
 
         let pointerHasMoved = (this._cursorX != x && this._cursorY != y);
-        let inWorkspace = (this.gridX < x && x < this.gridX + wsWidth &&
-                           this.gridY < y && y < this.gridY + wsHeight);
+        let inWorkspace = (this.x < x && x < this.x + wsWidth &&
+                           this.y < y && y < this.y + wsHeight);
 
         if (pointerHasMoved && inWorkspace) {
             // store current cursor position
@@ -1212,8 +996,11 @@ Workspace.prototype = {
     },
 
     showWindowsOverlays: function() {
+        if (this.leavingOverview)
+            return;
+
         this._windowOverlaysGroup.show();
-        this._fadeInAllOverlays();
+        this._showAllOverlays();
     },
 
     hideWindowsOverlays: function() {
@@ -1309,8 +1096,8 @@ Workspace.prototype = {
     },
 
     // check for maximized windows on the workspace
-    _haveMaximizedWindows: function() {
-        for (let i = 1; i < this._windows.length; i++) {
+    hasMaximizedWindows: function() {
+        for (let i = 0; i < this._windows.length; i++) {
             let metaWindow = this._windows[i].metaWindow;
             if (metaWindow.showing_on_its_workspace() &&
                 metaWindow.maximized_horizontally &&
@@ -1322,7 +1109,7 @@ Workspace.prototype = {
 
     // Animate the full-screen to Overview transition.
     zoomToOverview : function() {
-        this.actor.set_position(this.gridX, this.gridY);
+        this.actor.set_position(this.x, this.y);
         this.actor.set_scale(this.scale, this.scale);
 
         // Position and scale the windows.
@@ -1330,21 +1117,15 @@ Workspace.prototype = {
             this.positionWindows(WindowPositionFlags.ANIMATE | WindowPositionFlags.ZOOM);
         else
             this.positionWindows(WindowPositionFlags.ZOOM);
-
-        let active = global.screen.get_active_workspace();
-        let fadeInIcons = (Main.overview.animationInProgress &&
-                           active == this.metaWorkspace &&
-                           !this._haveMaximizedWindows());
-        this._desktop.zoomToOverview(fadeInIcons);
-
-        this._visible = true;
     },
 
     // Animates the return from Overview mode
     zoomFromOverview : function() {
+        let currentWorkspace = global.screen.get_active_workspace();
+
         this.leavingOverview = true;
 
-        this._hideAllOverlays();
+        this.hideWindowsOverlays();
 
         if (this._repositionWindowsId > 0) {
             Mainloop.source_remove(this._repositionWindowsId);
@@ -1353,8 +1134,11 @@ Workspace.prototype = {
         this._overviewHiddenId = Main.overview.connect('hidden', Lang.bind(this,
                                                                            this._doneLeavingOverview));
 
+        if (this._metaWorkspace == currentWorkspace)
+            return;
+
         // Position and scale the windows.
-        for (let i = 1; i < this._windows.length; i++) {
+        for (let i = 0; i < this._windows.length; i++) {
             let clone = this._windows[i];
 
             clone.zoomFromOverview();
@@ -1383,75 +1167,6 @@ Workspace.prototype = {
             }
         }
 
-        let active = global.screen.get_active_workspace();
-        let fadeOutIcons = (active == this.metaWorkspace &&
-                            !this._haveMaximizedWindows());
-        this._desktop.zoomFromOverview(fadeOutIcons);
-
-        this._visible = false;
-    },
-
-    // Animates grid shrinking/expanding when a row or column
-    // of workspaces is added or removed
-    resizeToGrid : function (oldScale) {
-        this._hideAllOverlays();
-        Tweener.addTween(this.actor,
-                         { x: this.gridX,
-                           y: this.gridY,
-                           scale_x: this.scale,
-                           scale_y: this.scale,
-                           time: Overview.ANIMATION_TIME,
-                           transition: 'easeOutQuad',
-                           onComplete: Lang.bind(this, this._fadeInAllOverlays)
-                         });
-    },
-
-    // Animates the addition of a new (empty) workspace
-    slideIn : function(oldScale) {
-        if (this.gridCol > this.gridRow) {
-            this.actor.set_position(global.screen_width, this.gridY);
-            this.actor.set_scale(oldScale, oldScale);
-        } else {
-            this.actor.set_position(this.gridX, global.screen_height);
-            this.actor.set_scale(this.scale, this.scale);
-        }
-        Tweener.addTween(this.actor,
-                         { x: this.gridX,
-                           y: this.gridY,
-                           scale_x: this.scale,
-                           scale_y: this.scale,
-                           time: Overview.ANIMATION_TIME,
-                           transition: 'easeOutQuad'
-                         });
-
-        this._visible = true;
-    },
-
-    // Animates the removal of a workspace
-    slideOut : function(onComplete) {
-        let destX = this.actor.x, destY = this.actor.y;
-
-        this._hideAllOverlays();
-
-        if (this.gridCol > this.gridRow)
-            destX = global.screen_width;
-        else
-            destY = global.screen_height;
-        Tweener.addTween(this.actor,
-                         { x: destX,
-                           y: destY,
-                           scale_x: this.scale,
-                           scale_y: this.scale,
-                           time: Overview.ANIMATION_TIME,
-                           transition: 'easeOutQuad',
-                           onComplete: onComplete
-                         });
-
-        this._visible = false;
-
-        // Don't let the user try to select this workspace as it's
-        // making its exit.
-        this._desktop.reactive = false;
     },
 
     destroy : function() {
@@ -1475,7 +1190,7 @@ Workspace.prototype = {
         // their parent (this.actor), but we might have a zoomed window
         // which has been reparented to the stage - _windows[0] holds
         // the desktop window, which is never reparented
-        for (let w = 1; w < this._windows.length; w++)
+        for (let w = 0; w < this._windows.length; w++)
             this._windows[w].destroy();
         this._windows = [];
     },
@@ -1487,8 +1202,7 @@ Workspace.prototype = {
 
     // Tests if @win belongs to this workspaces
     _isMyWindow : function (win) {
-        return win.get_workspace() == this.metaWorkspace.index() ||
-            (win.get_meta_window() && win.get_meta_window().is_on_all_workspaces());
+        return Main.isWindowActorDisplayedOnWorkspace(win, this.metaWorkspace.index());
     },
 
     // Tests if @win should be shown in the Overview
@@ -1506,12 +1220,12 @@ Workspace.prototype = {
                       Lang.bind(this, this._onCloneSelected));
         clone.connect('drag-begin',
                       Lang.bind(this, function(clone) {
-                          this.emit('window-drag-begin', clone.actor);
+                          Main.overview.beginWindowDrag();
                           overlay.hide();
                       }));
         clone.connect('drag-end',
                       Lang.bind(this, function(clone) {
-                          this.emit('window-drag-end', clone.actor);
+                          Main.overview.endWindowDrag();
                           overlay.show();
                       }));
         clone.connect('zoom-start',
@@ -1522,6 +1236,10 @@ Workspace.prototype = {
                       Lang.bind(this, function() {
                           this._windowIsZooming = false;
                       }));
+        clone.connect('size-changed',
+                      Lang.bind(this, function() {
+                          this.positionWindows(0);
+                      }));
 
         this.actor.add_actor(clone.actor);
 
@@ -1534,7 +1252,7 @@ Workspace.prototype = {
     },
 
     _onShowOverlayClose: function (windowOverlay) {
-        for (let i = 1; i < this._windowOverlays.length; i++) {
+        for (let i = 0; i < this._windowOverlays.length; i++) {
             let overlay = this._windowOverlays[i];
             if (overlay == windowOverlay)
                 continue;
@@ -1572,14 +1290,9 @@ Workspace.prototype = {
                             this.metaWorkspace.index());
     },
 
-    _removeSelf : function(actor, event) {
-        screen.remove_workspace(this.metaWorkspace, event.get_time());
-        return true;
-    },
-
     // Draggable target interface
     handleDragOver : function(source, actor, x, y, time) {
-        if (source instanceof WindowClone)
+        if (source.realWindow && !this._isMyWindow(source.realWindow))
             return DND.DragMotionResult.MOVE_DROP;
         if (source.shellWorkspaceLaunch)
             return DND.DragMotionResult.COPY_DROP;
@@ -1588,7 +1301,7 @@ Workspace.prototype = {
     },
 
     acceptDrop : function(source, actor, x, y, time) {
-        if (source instanceof WindowClone) {
+        if (source.realWindow) {
             let win = source.realWindow;
             if (this._isMyWindow(win))
                 return false;
@@ -1607,8 +1320,8 @@ Workspace.prototype = {
                                                  time);
             return true;
         } else if (source.shellWorkspaceLaunch) {
-            this.metaWorkspace.activate(time);
-            source.shellWorkspaceLaunch();
+            source.shellWorkspaceLaunch({ workspace: this.metaWorkspace,
+                                          timestamp: time });
             return true;
         }
 
@@ -1643,11 +1356,11 @@ function _workspaceRelativeModifier(workspace) {
     }
 
     return [ { name: 'x',
-               parameters: { workspacePos: workspace.gridX,
+               parameters: { workspacePos: workspace.x,
                              overviewPos: overviewPosX,
                              overviewScale: overviewScale } },
              { name: 'y',
-               parameters: { workspacePos: workspace.gridY,
+               parameters: { workspacePos: workspace.y,
                              overviewPos: overviewPosY,
                              overviewScale: overviewScale } }
            ];
