@@ -1228,9 +1228,7 @@ shell_global_gc (ShellGlobal *global)
 void
 shell_global_maybe_gc (ShellGlobal *global)
 {
-  JSContext *context = gjs_context_get_native_context (global->js_context);
-
-  JS_MaybeGC (context);
+  gjs_context_maybe_gc (global->js_context);
 }
 
 void
@@ -1630,6 +1628,7 @@ shell_global_get_current_time (ShellGlobal *global)
 {
   guint32 time;
   MetaDisplay *display;
+  const ClutterEvent *clutter_event;
 
   /* In case we have a xdnd timestamp use it */
   if (global->xdnd_timestamp != 0)
@@ -1652,8 +1651,16 @@ shell_global_get_current_time (ShellGlobal *global)
   time = meta_display_get_current_time (display);
   if (time != CLUTTER_CURRENT_TIME)
       return time;
+  /*
+   * We don't use clutter_get_current_event_time as it can give us a
+   * too old timestamp if there is no current event.
+   */
+  clutter_event = clutter_get_current_event ();
 
-  return clutter_get_current_event_time ();
+  if (clutter_event != NULL)
+    return clutter_event_get_time (clutter_event);
+  else
+    return CLUTTER_CURRENT_TIME;
 }
 
 /**
@@ -1891,6 +1898,7 @@ shell_global_run_at_leisure (ShellGlobal         *global,
 /**
  * shell_global_play_theme_sound:
  * @global: the #ShellGlobal
+ * @id: an id, used to cancel later (0 if not needed)
  * @name: the sound name
  *
  * Plays a simple sound picked according to Freedesktop sound theme.
@@ -1898,9 +1906,24 @@ shell_global_run_at_leisure (ShellGlobal         *global,
  */
 void
 shell_global_play_theme_sound (ShellGlobal *global,
+                               guint        id,
                                const char  *name)
 {
-  ca_context_play (global->sound_context, 0, CA_PROP_EVENT_ID, name, NULL);
+  ca_context_play (global->sound_context, id, CA_PROP_EVENT_ID, name, NULL);
+}
+
+/**
+ * shell_global_cancel_theme_sound:
+ * @global: the #ShellGlobal
+ * @id: the id previously passed to shell_global_play_theme_sound()
+ *
+ * Cancels a sound notification.
+ */
+void
+shell_global_cancel_theme_sound (ShellGlobal *global,
+                                 guint id)
+{
+  ca_context_cancel (global->sound_context, id);
 }
 
 /*
@@ -1966,3 +1989,63 @@ gboolean _shell_global_check_xdnd_event (ShellGlobal  *global,
 
     return FALSE;
 }
+
+/**
+ * ShellGetTpContactCb:
+ * @connection: The connection
+ * @contacts: (element-type TelepathyGLib.Contact): List of contacts
+ * @failed: Array of failed contacts
+ */
+
+static void
+shell_global_get_tp_contacts_cb (TpConnection *self,
+                                 guint n_contacts,
+                                 TpContact * const *contacts,
+                                 guint n_failed,
+                                 const TpHandle *failed,
+                                 const GError *error,
+                                 gpointer user_data,
+                                 GObject *weak_object)
+{
+  int i;
+  GList *contact_list = NULL;
+  for (i = 0; i < n_contacts; i++) {
+      contact_list = g_list_append(contact_list, contacts[i]);
+  }
+
+  TpHandle *failed_list = g_new0 (TpHandle, n_failed + 1);
+  memcpy(failed_list, failed, n_failed);
+
+  ((ShellGetTpContactCb)user_data)(self, contact_list, failed_list);
+}
+
+
+/**
+ * shell_get_tp_contacts:
+ * @self: A connection, which must be ready
+ * @n_handles: Number of handles in handles
+ * @handles: (array length=n_handles) (element-type uint): Array of handles
+ * @n_features: Number of features in features
+ * @features: (array length=n_features) (allow-none) (element-type uint):
+ *  Array of features
+ * @callback: (scope async): User callback to run when the contacts are ready
+ *
+ * Wrap tp_connection_get_contacts_by_handle so we can transform the array
+ * into a null-terminated one, which gjs can handle.
+ * We send the original callback to tp_connection_get_contacts_by_handle as
+ * user_data, and we have our own function as callback, which does the
+ * transforming.
+ */
+void
+shell_get_tp_contacts (TpConnection *self,
+                       guint n_handles,
+                       const TpHandle *handles,
+                       guint n_features,
+                       const TpContactFeature *features,
+                       ShellGetTpContactCb callback)
+{
+  tp_connection_get_contacts_by_handle(self, n_handles, handles,
+                                       n_features, features,
+                                       shell_global_get_tp_contacts_cb,
+                                       callback, NULL, NULL);
+}
