@@ -599,31 +599,6 @@ st_entry_key_focus_in (ClutterActor *actor)
   clutter_actor_grab_key_focus (priv->entry);
 }
 
-static gboolean
-st_entry_enter_event (ClutterActor         *actor,
-                      ClutterCrossingEvent *event)
-{
-  StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
-
-  if (clutter_actor_contains (actor, event->source)
-      && priv->hint && priv->hint_visible)
-    {
-      st_widget_set_hover (ST_WIDGET (actor), TRUE);
-    }
-
-  return CLUTTER_ACTOR_CLASS (st_entry_parent_class)->enter_event (actor, event);
-}
-
-static gboolean
-st_entry_leave_event (ClutterActor         *actor,
-                      ClutterCrossingEvent *event)
-{
-  if (!clutter_actor_contains (actor, event->related))
-    st_widget_set_hover (ST_WIDGET (actor), FALSE);
-
-  return CLUTTER_ACTOR_CLASS (st_entry_parent_class)->leave_event (actor, event);
-}
-
 static void
 st_entry_class_init (StEntryClass *klass)
 {
@@ -649,8 +624,6 @@ st_entry_class_init (StEntryClass *klass)
 
   actor_class->key_press_event = st_entry_key_press_event;
   actor_class->key_focus_in = st_entry_key_focus_in;
-  actor_class->enter_event = st_entry_enter_event;
-  actor_class->leave_event = st_entry_leave_event;
 
   widget_class->style_changed = st_entry_style_changed;
   widget_class->navigate_focus = st_entry_navigate_focus;
@@ -895,9 +868,9 @@ _st_entry_icon_press_cb (ClutterActor       *actor,
 }
 
 static void
-_st_entry_set_icon_from_file (StEntry       *entry,
-                              ClutterActor **icon,
-                              const gchar   *filename)
+_st_entry_set_icon (StEntry       *entry,
+                    ClutterActor **icon,
+                    ClutterActor  *new_icon)
 {
   if (*icon)
     {
@@ -908,13 +881,9 @@ _st_entry_set_icon_from_file (StEntry       *entry,
       *icon = NULL;
     }
 
-  if (filename)
+  if (new_icon)
     {
-      StTextureCache *cache;
-
-      cache = st_texture_cache_get_default ();
-
-      *icon = (ClutterActor*) st_texture_cache_load_file_simple (cache, filename);
+      *icon = g_object_ref (new_icon);
 
       clutter_actor_set_reactive (*icon, TRUE);
       clutter_actor_set_parent (*icon, CLUTTER_ACTOR (entry));
@@ -925,6 +894,65 @@ _st_entry_set_icon_from_file (StEntry       *entry,
   clutter_actor_queue_relayout (CLUTTER_ACTOR (entry));
 }
 
+static void
+_st_entry_set_icon_from_file (StEntry       *entry,
+                              ClutterActor **icon,
+                              const gchar   *filename)
+{
+  ClutterActor *new_icon = NULL;
+
+  if (filename)
+    {
+      StTextureCache *cache;
+
+      cache = st_texture_cache_get_default ();
+
+      new_icon = (ClutterActor*) st_texture_cache_load_file_simple (cache, filename);
+    }
+
+  _st_entry_set_icon  (entry, icon, new_icon);
+}
+
+/**
+ * st_entry_set_primary_icon:
+ * @entry: a #StEntry
+ * @icon: (allow-none): a #ClutterActor
+ *
+ * Set the primary icon of the entry to @icon
+ */
+void
+st_entry_set_primary_icon (StEntry      *entry,
+                           ClutterActor *icon)
+{
+  StEntryPrivate *priv;
+
+  g_return_if_fail (ST_IS_ENTRY (entry));
+
+  priv = entry->priv;
+
+  _st_entry_set_icon (entry, &priv->primary_icon, icon);
+}
+
+/**
+ * st_entry_set_secondary_icon:
+ * @entry: a #StEntry
+ * @icon: (allow-none): an #ClutterActor
+ *
+ * Set the secondary icon of the entry to @icon
+ */
+void
+st_entry_set_secondary_icon (StEntry      *entry,
+                             ClutterActor *icon)
+{
+  StEntryPrivate *priv;
+
+  g_return_if_fail (ST_IS_ENTRY (entry));
+
+  priv = entry->priv;
+
+  _st_entry_set_icon (entry, &priv->secondary_icon, icon);
+}
+
 /**
  * st_entry_set_primary_icon_from_file:
  * @entry: a #StEntry
