@@ -4,6 +4,8 @@
  *
  * Copyright 2008 OpenedHand
  * Copyright 2009 Intel Corporation.
+ * Copyright 2009, 2010 Red Hat, Inc.
+ * Copyright 2010 Maxim Ermilov
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU Lesser General Public License,
@@ -15,12 +17,7 @@
  * more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
- *
- * Written by: Chris Lord <chris@openedhand.com>
- * Port to St by: Robert Staudinger <robsta@openedhand.com>
- *
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
  */
 
 /**
@@ -65,6 +62,7 @@
 #include "st-marshal.h"
 #include "st-scroll-bar.h"
 #include "st-scrollable.h"
+#include "st-scroll-view-fade.h"
 #include <clutter/clutter.h>
 #include <math.h>
 
@@ -80,9 +78,6 @@ G_DEFINE_TYPE_WITH_CODE (StScrollView, st_scroll_view, ST_TYPE_BIN,
                                                              ST_TYPE_SCROLL_VIEW, \
                                                              StScrollViewPrivate))
 
-/* Default width (or height - the narrow dimension) for the scrollbars*/
-#define DEFAULT_SCROLLBAR_WIDTH 24
-
 struct _StScrollViewPrivate
 {
   /* a pointer to the child; this is actually stored
@@ -99,20 +94,17 @@ struct _StScrollViewPrivate
   GtkPolicyType hscrollbar_policy;
   GtkPolicyType vscrollbar_policy;
 
-  ClutterActor *top_shadow;
-  ClutterActor *bottom_shadow;
-
   gfloat        row_size;
   gfloat        column_size;
 
-  gboolean      vshadows;
+  gboolean      vfade;
+  StScrollViewFade *vfade_effect;
+
   gboolean      row_size_set : 1;
   gboolean      column_size_set : 1;
   guint         mouse_scroll : 1;
   guint         hscrollbar_visible : 1;
   guint         vscrollbar_visible : 1;
-  guint         top_shadow_visible : 1;
-  guint         bottom_shadow_visible : 1;
 };
 
 enum {
@@ -123,7 +115,7 @@ enum {
   PROP_HSCROLLBAR_POLICY,
   PROP_VSCROLLBAR_POLICY,
   PROP_MOUSE_SCROLL,
-  PROP_VSHADOWS
+  PROP_VFADE
 };
 
 static void
@@ -151,80 +143,50 @@ st_scroll_view_get_property (GObject    *object,
     case PROP_MOUSE_SCROLL:
       g_value_set_boolean (value, priv->mouse_scroll);
       break;
-    case PROP_VSHADOWS:
-      g_value_set_boolean (value, priv->vshadows);
+    case PROP_VFADE:
+      g_value_set_boolean (value, priv->vfade);
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
 }
 
-static void
-update_shadow_visibility (StScrollView *scroll)
-{
-  StScrollViewPrivate *priv = scroll->priv;
-
-  if (priv->vshadows)
-    {
-      gdouble value, lower, upper, page_size;
-
-      st_adjustment_get_values (priv->vadjustment, &value, &lower, &upper, NULL, NULL, &page_size);
-
-      priv->top_shadow_visible = value > lower + 0.1;
-      priv->bottom_shadow_visible = value < upper - page_size - 0.1;
-    }
-  else
-    {
-      priv->top_shadow_visible = FALSE;
-      priv->bottom_shadow_visible = FALSE;
-    }
-}
-
 /**
- * st_scroll_view_set_vshadows:
+ * st_scroll_view_set_vfade:
  * @self: a #StScrollView
- * @vshadows: Whether to enable vertical shadows
+ * @vfade: Whether to enable the vertical fade effect
  *
- * Sets whether to show shadows at the top and bottom of the area. Shadows
- * are omitted when fully scrolled to that edge.
+ * Sets whether to fade the content at the top and bottom of the area when not
+ * fully scrolled to that edge.
  */
 void
-st_scroll_view_set_vshadows (StScrollView *self,
-                             gboolean      vshadows)
+st_scroll_view_set_vfade (StScrollView *self,
+                          gboolean vfade)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (self)->priv;
 
-  vshadows = vshadows != FALSE;
-  if (priv->vshadows == vshadows)
+  vfade = vfade != FALSE;
+  if (priv->vfade == vfade)
     return;
 
-  priv->vshadows = vshadows;
+  priv->vfade = vfade;
 
-  if (vshadows)
+  if (vfade)
     {
-      if (priv->top_shadow)
-        {
-          clutter_actor_show (priv->top_shadow);
-          clutter_actor_show (priv->bottom_shadow);
-        }
-      else
-        {
-          priv->top_shadow = g_object_new (ST_TYPE_BIN, "style-class", "top-shadow", NULL);
-          priv->bottom_shadow = g_object_new (ST_TYPE_BIN, "style-class", "bottom-shadow", NULL);
+      if (priv->vfade_effect == NULL)
+        priv->vfade_effect = g_object_new (ST_TYPE_SCROLL_VIEW_FADE, NULL);
 
-          clutter_actor_set_parent (priv->bottom_shadow, CLUTTER_ACTOR (self));
-          clutter_actor_set_parent (priv->top_shadow, CLUTTER_ACTOR (self));
-        }
+      clutter_actor_add_effect (CLUTTER_ACTOR (self), CLUTTER_EFFECT (priv->vfade_effect));
     }
-  else
+   else
     {
-      clutter_actor_hide (priv->top_shadow);
-      clutter_actor_hide (priv->bottom_shadow);
+      clutter_actor_remove_effect (CLUTTER_ACTOR (self), CLUTTER_EFFECT (priv->vfade_effect));
+      priv->vfade_effect = NULL;
     }
 
-  update_shadow_visibility (self);
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (self));
 
-  g_object_notify (G_OBJECT (self), "vshadows");
+  g_object_notify (G_OBJECT (self), "vfade");
 }
 
 static void
@@ -238,8 +200,8 @@ st_scroll_view_set_property (GObject      *object,
 
   switch (property_id)
     {
-    case PROP_VSHADOWS:
-      st_scroll_view_set_vshadows (self, g_value_get_boolean (value));
+    case PROP_VFADE:
+      st_scroll_view_set_vfade (self, g_value_get_boolean (value));
       break;
     case PROP_MOUSE_SCROLL:
       st_scroll_view_set_mouse_scrolling (self,
@@ -265,6 +227,12 @@ st_scroll_view_dispose (GObject *object)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (object)->priv;
 
+  if (priv->vfade_effect)
+    {
+      clutter_actor_remove_effect (CLUTTER_ACTOR (object), CLUTTER_EFFECT (priv->vfade_effect));
+      priv->vfade_effect = NULL;
+    }
+
   if (priv->vscroll)
     clutter_actor_destroy (priv->vscroll);
 
@@ -287,21 +255,7 @@ st_scroll_view_dispose (GObject *object)
     {
       g_object_run_dispose (G_OBJECT (priv->vadjustment));
       g_object_unref (priv->vadjustment);
-      priv->hadjustment = NULL;
-    }
-
-  /* since it's impossible to get a handle to these actors, we can
-   * just directly unparent them and not go through destroy/remove */
-  if (priv->top_shadow)
-    {
-      clutter_actor_unparent (priv->top_shadow);
-      priv->top_shadow = NULL;
-    }
-
-  if (priv->bottom_shadow)
-    {
-      clutter_actor_unparent (priv->bottom_shadow);
-      priv->bottom_shadow = NULL;
+      priv->vadjustment = NULL;
     }
 
   G_OBJECT_CLASS (st_scroll_view_parent_class)->dispose (object);
@@ -320,11 +274,6 @@ st_scroll_view_paint (ClutterActor *actor)
     clutter_actor_paint (priv->hscroll);
   if (priv->vscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
     clutter_actor_paint (priv->vscroll);
-
-  if (priv->top_shadow_visible)
-    clutter_actor_paint (priv->top_shadow);
-  if (priv->bottom_shadow_visible)
-    clutter_actor_paint (priv->bottom_shadow);
 }
 
 static void
@@ -344,36 +293,37 @@ st_scroll_view_pick (ClutterActor       *actor,
 }
 
 static double
-get_scrollbar_width (StScrollView *scroll)
+get_scrollbar_width (StScrollView *scroll,
+                     gfloat        for_height)
 {
   StScrollViewPrivate *priv = scroll->priv;
 
   if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
     {
-      StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll));
-      double result = DEFAULT_SCROLLBAR_WIDTH;
+      gfloat min_size;
 
-      st_theme_node_get_length (theme_node, "scrollbar-width", FALSE, &result);
-
-      return result;
+      clutter_actor_get_preferred_width (CLUTTER_ACTOR (priv->vscroll), for_height,
+                                         &min_size, NULL);
+      return min_size;
     }
   else
     return 0;
 }
 
 static double
-get_scrollbar_height (StScrollView *scroll)
+get_scrollbar_height (StScrollView *scroll,
+                      gfloat        for_width)
 {
   StScrollViewPrivate *priv = scroll->priv;
 
   if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
     {
-      StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (scroll));
-      double result = DEFAULT_SCROLLBAR_WIDTH;
+      gfloat min_size;
 
-      st_theme_node_get_length (theme_node, "scrollbar-height", FALSE, &result);
+      clutter_actor_get_preferred_height (CLUTTER_ACTOR (priv->hscroll), for_width,
+                                          &min_size, NULL);
 
-      return result;
+      return min_size;
     }
   else
     return 0;
@@ -434,7 +384,7 @@ st_scroll_view_get_preferred_width (ClutterActor *actor,
   account_for_vscrollbar = priv->vscrollbar_policy != GTK_POLICY_NEVER;
   if (account_for_vscrollbar)
     {
-      float sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
+      float sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor), for_height);
 
       min_width += sb_width;
       natural_width += sb_width;
@@ -476,7 +426,7 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
   if (min_height_p)
     *min_height_p = 0;
 
-  sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
+  sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor), -1);
 
   switch (priv->vscrollbar_policy)
     {
@@ -522,7 +472,7 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
 
   if (account_for_hscrollbar)
     {
-      float sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor));
+      float sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor), for_width);
 
       min_height += sb_height;
       natural_height += sb_height;
@@ -537,18 +487,6 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
   st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
 }
 
-static gfloat
-get_shadow_height (ClutterActor *shadow)
-{
-  gfloat natural_height;
-
-  /* The shadows are empty StBin and have no height-for-width behavior */
-
-  clutter_actor_get_preferred_height (shadow, -1, NULL, &natural_height);
-
-  return natural_height;
-}
-
 static void
 st_scroll_view_allocate (ClutterActor          *actor,
                          const ClutterActorBox *box,
@@ -574,14 +512,21 @@ st_scroll_view_allocate (ClutterActor          *actor,
   CLUTTER_ACTOR_CLASS (parent_parent_class)->
   allocate (actor, box, flags);
 
-
   st_theme_node_get_content_box (theme_node, box, &content_box);
 
   avail_width = content_box.x2 - content_box.x1;
   avail_height = content_box.y2 - content_box.y1;
 
-  sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor));
-  sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor));
+  if (clutter_actor_get_request_mode (actor) == CLUTTER_REQUEST_HEIGHT_FOR_WIDTH)
+    {
+      sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor), -1);
+      sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor), sb_width);
+    }
+  else
+    {
+      sb_height = get_scrollbar_height (ST_SCROLL_VIEW (actor), -1);
+      sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor), sb_height);
+    }
 
   /* Determine what scrollbars are visible. The basic idea of the
    * handling of an automatic scrollbars is that we start off with the
@@ -686,25 +631,6 @@ st_scroll_view_allocate (ClutterActor          *actor,
   if (priv->child)
     clutter_actor_allocate (priv->child, &child_box, flags);
 
-  /* Shadows */
-  if (priv->top_shadow && CLUTTER_ACTOR_IS_VISIBLE (priv->top_shadow))
-    {
-      child_box.x1 = content_box.x1;
-      child_box.y1 = content_box.y1;
-      child_box.x2 = MAX (child_box.x1, content_box.x2 - sb_width);
-      child_box.y2 = content_box.y1 + get_shadow_height (priv->top_shadow);
-      clutter_actor_allocate (priv->top_shadow, &child_box, flags);
-    }
-
-  if (priv->bottom_shadow && CLUTTER_ACTOR_IS_VISIBLE (priv->bottom_shadow))
-    {
-      child_box.x1 = content_box.x1;
-      child_box.y1 = content_box.y2 - sb_height - get_shadow_height (priv->bottom_shadow);
-      child_box.x2 = MAX (content_box.x1, content_box.x2 - sb_width);
-      child_box.y2 = content_box.y2 - sb_height;
-      clutter_actor_allocate (priv->bottom_shadow, &child_box, flags);
-    }
-
   priv->hscrollbar_visible = hscrollbar_visible;
   priv->vscrollbar_visible = vscrollbar_visible;
 }
@@ -717,12 +643,6 @@ st_scroll_view_style_changed (StWidget *widget)
   st_widget_style_changed (ST_WIDGET (priv->hscroll));
   st_widget_style_changed (ST_WIDGET (priv->vscroll));
 
-  if (priv->top_shadow)
-    {
-      st_widget_style_changed (ST_WIDGET (priv->top_shadow));
-      st_widget_style_changed (ST_WIDGET (priv->bottom_shadow));
-    }
-
   ST_WIDGET_CLASS (st_scroll_view_parent_class)->style_changed (widget);
 }
 
@@ -855,31 +775,16 @@ st_scroll_view_class_init (StScrollViewClass *klass)
                                    PROP_MOUSE_SCROLL,
                                    pspec);
 
-  pspec = g_param_spec_boolean ("vshadows",
+  pspec = g_param_spec_boolean ("vfade",
                                 "Vertical Shadows",
-                                "Show shadows at the top and and bottom of the area unless fully scrolled to that edge",
+                                "Fade the content at the top and and bottom of the area unless fully scrolled to that edge",
                                 FALSE,
                                 G_PARAM_READWRITE);
   g_object_class_install_property (object_class,
-                                   PROP_VSHADOWS,
+                                   PROP_VFADE,
                                    pspec);
 }
 
-static void
-child_adjustment_changed_cb (StAdjustment *adjustment,
-                             StScrollView *scroll)
-{
-  update_shadow_visibility (scroll);
-}
-
-static void
-child_adjustment_notify_value (GObject      *gobject,
-                               GParamSpec   *pspec,
-                               StScrollView *scroll)
-{
-  update_shadow_visibility (scroll);
-}
-
 static void
 st_scroll_view_init (StScrollView *self)
 {
@@ -895,10 +800,6 @@ st_scroll_view_init (StScrollView *self)
                                 NULL);
 
   priv->vadjustment = g_object_new (ST_TYPE_ADJUSTMENT, NULL);
-  g_signal_connect (priv->vadjustment, "changed",
-                    G_CALLBACK (child_adjustment_changed_cb), self);
-  g_signal_connect (priv->vadjustment, "notify::value",
-                    G_CALLBACK (child_adjustment_notify_value), self);
   priv->vscroll = g_object_new (ST_TYPE_SCROLL_BAR,
                                 "adjustment", priv->vadjustment,
                                 "vertical", TRUE,
@@ -986,12 +887,6 @@ st_scroll_view_foreach_with_internals (ClutterContainer *container,
 
   if (priv->vscroll != NULL)
     callback (priv->vscroll, user_data);
-
-  if (priv->top_shadow)
-    {
-      callback (priv->top_shadow, user_data);
-      callback (priv->bottom_shadow, user_data);
-    }
 }
 
 static void
