@@ -30,6 +30,7 @@
 
 #define CACHE_PREFIX_GICON "gicon:"
 #define CACHE_PREFIX_URI "uri:"
+#define CACHE_PREFIX_URI_FOR_CAIRO "uri-for-cairo:"
 #define CACHE_PREFIX_THUMBNAIL_URI "thumbnail-uri:"
 #define CACHE_PREFIX_RAW_CHECKSUM "raw-checksum:"
 #define CACHE_PREFIX_COMPRESSED_CHECKSUM "compressed-checksum:"
@@ -437,6 +438,67 @@ out:
   return rotated_pixbuf;
 }
 
+static GdkPixbuf*
+decode_image (const char *val)
+{
+  int i;
+  GError *error = NULL;
+  GdkPixbuf *res = NULL;
+  struct {
+    const char *prefix;
+    const char *mime_type;
+  } formats[] = {
+    { "data:image/x-icon;base64,", "image/x-icon" },
+    { "data:image/png;base64,", "image/png" }
+  };
+
+  g_return_val_if_fail (val, NULL);
+
+  for (i = 0; i < G_N_ELEMENTS (formats); i++)
+    {
+      if (g_str_has_prefix (val, formats[i].prefix))
+        {
+          gsize len;
+          guchar *data = NULL;
+          char *unescaped;
+
+          unescaped = g_uri_unescape_string (val + strlen (formats[i].prefix), NULL);
+          if (unescaped)
+            {
+              data = g_base64_decode (unescaped, &len);
+              g_free (unescaped);
+            }
+
+          if (data)
+            {
+              GdkPixbufLoader *loader;
+
+              loader = gdk_pixbuf_loader_new_with_mime_type (formats[i].mime_type, &error);
+              if (loader &&
+                  gdk_pixbuf_loader_write (loader, data, len, &error) &&
+                  gdk_pixbuf_loader_close (loader, &error))
+                {
+                  res = gdk_pixbuf_loader_get_pixbuf (loader);
+                  g_object_ref (res);
+                }
+              g_object_unref (loader);
+              g_free (data);
+            }
+        }
+    }
+  if (!res)
+    {
+      if (error)
+        {
+          g_warning ("%s\n", error->message);
+          g_error_free (error);
+        }
+      else
+        g_warning ("incorrect data uri");
+    }
+  return res;
+}
+
 static GdkPixbuf *
 impl_load_pixbuf_file (const char     *uri,
                        int             available_width,
@@ -448,6 +510,9 @@ impl_load_pixbuf_file (const char     *uri,
   char *contents = NULL;
   gsize size;
 
+  if (g_str_has_prefix (uri, "data:"))
+    return decode_image (uri);
+
   file = g_file_new_for_uri (uri);
   if (g_file_load_contents (file, NULL, &contents, &size, NULL, error))
     {
@@ -734,6 +799,27 @@ pixbuf_to_cogl_handle (GdkPixbuf *pixbuf)
                                      gdk_pixbuf_get_pixels (pixbuf));
 }
 
+static cairo_surface_t *
+pixbuf_to_cairo_surface (GdkPixbuf *pixbuf)
+{
+  cairo_surface_t *dummy_surface;
+  cairo_pattern_t *pattern;
+  cairo_surface_t *surface;
+  cairo_t *cr;
+
+  dummy_surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 1, 1);
+
+  cr = cairo_create (dummy_surface);
+  gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
+  pattern = cairo_get_source (cr);
+  cairo_pattern_get_surface (pattern, &surface);
+  cairo_surface_reference (surface);
+  cairo_destroy (cr);
+  cairo_surface_destroy (dummy_surface);
+
+  return surface;
+}
+
 static GdkPixbuf *
 load_pixbuf_fallback(AsyncTextureLoadData *data)
 {
@@ -1183,9 +1269,7 @@ on_sliced_image_loaded (GObject *source_object,
     {
       ClutterActor *actor = load_from_pixbuf (GDK_PIXBUF (list->data));
       clutter_actor_hide (actor);
-
       clutter_container_add_actor (CLUTTER_CONTAINER (data->group), actor);
-      g_object_unref (list->data);
     }
 }
 
@@ -1198,6 +1282,17 @@ on_data_destroy (gpointer data)
   g_free (d);
 }
 
+static void
+free_glist_unref_gobjects (gpointer p)
+{
+  GList *list = p;
+  GList *iter;
+
+  for (iter = list; iter; iter = iter->next)
+    g_object_unref (iter->data);
+  g_list_free (list);
+}
+
 static void
 load_sliced_image (GSimpleAsyncResult *result,
                    GObject *object,
@@ -1223,17 +1318,14 @@ load_sliced_image (GSimpleAsyncResult *result,
       for (x = 0; x < width; x += data->grid_height)
         {
           GdkPixbuf *pixbuf = gdk_pixbuf_new_subpixbuf (pix, x, y, data->grid_width, data->grid_height);
-          if (!pixbuf)
-            {
-              g_simple_async_result_set_error (result, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
-                                               "Has failed thumbnail");
-              break;
-            }
+          g_assert (pixbuf != NULL);
           res = g_list_append (res, pixbuf);
         }
     }
-  if (res)
-    g_simple_async_result_set_op_res_gpointer (result, res, (GDestroyNotify)g_list_free);
+  /* We don't need the original pixbuf anymore, though the subpixbufs
+     will hold a reference. */
+  g_object_unref (pix);
+  g_simple_async_result_set_op_res_gpointer (result, res, free_glist_unref_gobjects);
 }
 
 /**
@@ -1434,6 +1526,45 @@ out:
   return texdata;
 }
 
+static cairo_surface_t *
+st_texture_cache_load_uri_sync_to_cairo_surface (StTextureCache        *cache,
+                                                 StTextureCachePolicy   policy,
+                                                 const gchar           *uri,
+                                                 int                    available_width,
+                                                 int                    available_height,
+                                                 GError               **error)
+{
+  cairo_surface_t *surface;
+  GdkPixbuf *pixbuf;
+  char *key;
+
+  key = g_strconcat (CACHE_PREFIX_URI_FOR_CAIRO, uri, NULL);
+
+  surface = g_hash_table_lookup (cache->priv->keyed_cache, key);
+
+  if (surface == NULL)
+    {
+      pixbuf = impl_load_pixbuf_file (uri, available_width, available_height, error);
+      if (!pixbuf)
+        goto out;
+
+      surface = pixbuf_to_cairo_surface (pixbuf);
+      g_object_unref (pixbuf);
+
+      if (policy == ST_TEXTURE_CACHE_POLICY_FOREVER)
+        {
+          cairo_surface_reference (surface);
+          g_hash_table_insert (cache->priv->keyed_cache, g_strdup (key), surface);
+        }
+    }
+  else
+    cairo_surface_reference (surface);
+
+out:
+  g_free (key);
+  return surface;
+}
+
 /**
  * st_texture_cache_load_uri_sync:
  *
@@ -1512,6 +1643,43 @@ st_texture_cache_load_file_to_cogl_texture (StTextureCache *cache,
   return texture;
 }
 
+/**
+ * st_texture_cache_load_file_to_cairo_surface:
+ * @cache: A #StTextureCache
+ * @file_path: Path to a file in supported image format
+ *
+ * This function synchronously loads the given file path
+ * into a cairo surface.  On error, a warning is emitted
+ * and %NULL is returned.
+ *
+ * Returns: (transfer full): a new #cairo_surface_t *
+ */
+cairo_surface_t *
+st_texture_cache_load_file_to_cairo_surface (StTextureCache *cache,
+                                             const gchar    *file_path)
+{
+  cairo_surface_t *surface;
+  GFile *file;
+  char *uri;
+  GError *error = NULL;
+
+  file = g_file_new_for_path (file_path);
+  uri = g_file_get_uri (file);
+
+  surface = st_texture_cache_load_uri_sync_to_cairo_surface (cache, ST_TEXTURE_CACHE_POLICY_FOREVER,
+                                                             uri, -1, -1, &error);
+  g_object_unref (file);
+  g_free (uri);
+
+  if (surface == NULL)
+    {
+      g_warning ("Failed to load %s: %s", file_path, error->message);
+      g_clear_error (&error);
+      return NULL;
+    }
+  return surface;
+}
+
 /**
  * st_texture_cache_load_file_simple:
  * @cache: A #StTextureCache
@@ -1609,7 +1777,7 @@ st_texture_cache_load_from_data (StTextureCache    *cache,
 /**
  * st_texture_cache_load_from_raw:
  * @cache: a #StTextureCache
- * @data: raw pixel data
+ * @data: (array length=len): raw pixel data
  * @len: the length of @data
  * @has_alpha: whether @data includes an alpha channel
  * @width: width in pixels of @data
