@@ -444,6 +444,9 @@ st_widget_unmap (ClutterActor *actor)
 
   if (priv->tooltip)
     clutter_actor_unmap ((ClutterActor *) priv->tooltip);
+
+  if (priv->track_hover && priv->hover)
+    st_widget_set_hover (ST_WIDGET (actor), FALSE);
 }
 
 static void notify_children_of_style_change (ClutterContainer *container);
@@ -565,7 +568,8 @@ st_widget_get_theme_node (StWidget *widget)
 
       if (stage == NULL)
         {
-          g_error ("st_widget_get_theme_node called on a widget not in a stage");
+          g_error ("st_widget_get_theme_node called on the widget %s which is not in the stage.",
+                    st_describe_actor (CLUTTER_ACTOR (widget)));
         }
 
       if (parent_node == NULL)
